import enum
import logging
from dataclasses import dataclass

from PIL import Image

from .util import GpuObject, BindSemanticObject
from .. import gl


class TextureKind(enum.Enum):
    # The texture is combined with the result of the diffuse lighting equation.
    DiffuseMap = 0
    # The texture is combined with the result of the specular lighting equation.
    SpecularMap = 1
    # The texture is combined with the result of the ambient lighting equation.
    AmbientMap = 2
    # The texture is added to the result of the lighting calculation.It isn't influenced by incoming light.
    EmissiveMap = 3
    # The texture is a height map.
    HeightMap = 4
    # The texture is a bump map.
    BumpMap = 5
    # The texture is a(tangent space) normal - map.
    NormalMap = 6
    # The texture defines the glossiness of the material.
    ShininessMap = 7
    # The texture defines per - pixel opacity.
    OpacityMap = 8
    # Displacement texture
    DisplacementMap = 9
    # Lightmap texture(aka Ambient Occlusion)
    LightMap = 10
    # Reflection texture.
    ReflectionMap = 11


class TextureWrapMode(enum.Enum):
    # Tiles the texture, creating a repeating pattern (repeat wrapping).
    Repeat = gl.GL_REPEAT
    # Clamps the texture to the last pixel at the edge (clamp to edge wrapping).
    ClampEdge = gl.GL_CLAMP_TO_EDGE
    # Tiles the texture, creating a repeating pattern by mirroring it at every integer
    # boundary (mirrored repeat wrapping).
    Mirror = gl.GL_MIRRORED_REPEAT
    ClampBorder = gl.GL_CLAMP_TO_BORDER


class FilterMode(enum.Enum):
    Linear = gl.GL_LINEAR
    Nearest = gl.GL_NEAREST


@dataclass(frozen=True)
class ObjTextureParams:
    PARAM_NAMES = ['o', 's', 't', 'bm', 'blendu', 'blendv', 'cc', 'clamp', 'texres', 'mm', 'imfchan', 'type']
    """
    Parameters for a texture loaded from an OBJ file.
    """
    # Offsets added to the u, v and w coordinates of the texture map.
    # Default is (0, 0, 0).
    #
    # -o u v w
    offset: tuple = (0.0, 0.0, 0.0)
    # Scaling factors applied to the u, v and w coordinates of the texture map.
    # Default is (1, 1, 1).
    #
    # -s u v w
    scale: tuple = (1.0, 1.0, 1.0)
    # Specifies the turbulence values. Adding turbulence to a texture along a
    # specified direction adds variance to the original image and allows a simple
    # image to be repeated over a larger area without noticeable tiling effects.
    # The default is (0, 0, 0).
    #
    # -t u v w
    turbulence: tuple = (0.0, 0.0, 0.0)
    # Bias and gain values used to modify the color values of the texture map.
    # Modifies the range over which scalar or color texture values may vary. This
    # has an effect only during rendering and does not change the file.
    #
    # "base" adds a base value to the texture values. Default is 0.0. A positive
    # value makes everything brighter, a negative value makes everything dimmer.
    #
    # "gain" expands the range of the texture values. Default is 1.0; the range is
    # unlimited. Increasing the number increases the contrast.
    #
    # -mm base gain
    mm = (0.0, 1.0)
    # Turns on or off the clamping of texture coordinates. When clamping is on,
    # texture coordinates are limited to the range [0, 1]. Default is off.
    # When clamping is turned on, one copy of the texture is mapped onto the
    # surface, rather than repeating copies of the original texture across the
    # surface of a polygon, which is the default.  Outside the origin
    # texture, the underlying material is unchanged.
    #
    # -clamp on | off
    clamp = False
    # Turns on or off the texture blending in the vertical direction. Default is on.
    #
    # -blendu on | off
    blend_u = True
    # Turns on or off the texture blending in the horizontal direction. Default is on.
    #
    # -blendv on | off
    blend_v = True
    # Specifies a multiplier for the bump map values. Values stored with the texture are
    # multiplied by this value before they are applied to the surface. The default value is 1.0.
    # It can be positive or negative. For best results, the value should be between -1.0 and 1.0.
    #
    # -bm mult
    bump_mult = 1.0
    # Specifies the channel of the texture used to create a scalar or bump texture. Scalar
    # textures are applied to transparency, specular exponent, decal, and displacement maps.
    # Channel values can be r, g, b, m, l, or z. The default for bump and scalar textures is
    # 'l'(luminance), unless you are building a decal. In that case, the default is 'm'(matte).
    #
    # -imfchan r | g | b | m | l | z
    imfchan = 'l'
    # Turns on or off the color correction for the texture mao. It can only be used with
    # map_Ka, map_Kd, and map_Ks. Default is off.
    #
    # -cc on | off
    color_correction = False
    # Specifies the resolution of texture created when and image is used. The default texture
    # size is the largest power of two that does not exceed the original image size.
    #
    # If the source image is an exact power of 2, the texture cannot be built
    # any larger.  If the source image size is not an exact power of 2, you
    # can specify that the texture be built at the next power of 2 greater
    # than the source image size.
    #
    # The original image should be square, otherwise, it will be scaled to
    # fit the closest square size that is not larger than the original.
    # Scaling reduces sharpness.
    #
    # -texres resolution
    tex_res = None
    # Used in combination with reflection map, to specify how the image should be interpreted.
    # It can be 'cube_top', 'cube_bottom', 'cube_left', 'cube_right', 'cube_front', 'cube_back', 'sphere'.
    #
    # refl -type sphere -options -args filename
    refl_type = None


class Texture(GpuObject, BindSemanticObject):
    def __init__(self, image_path, image, kind=TextureKind.DiffuseMap, target=gl.GL_TEXTURE_2D,
                 wrap_mode=TextureWrapMode.Repeat, filter_mode=FilterMode.Linear, gen_mipmap=True, obj_tex_params=None):
        super().__init__(gl.GL_TEXTURE_2D, -1)
        self._wrap_mode = wrap_mode
        self._filter_mode = filter_mode
        self._name_path = image_path
        self._target = target
        self._image = image
        self._kind = kind
        num_comps = self._image.num_channels
        self._format = gl.GL_RED if num_comps == 1 else gl.GL_RGB if num_comps == 3 else gl.GL_RGBA

        self._id = gl.glGenTextures(1)
        gl.glBindTexture(self._target, self._id)
        gl.glTexImage2D(self._target, 0, self._format, self._image.width, self._image.height,
                        0, self._format, gl.GL_UNSIGNED_BYTE, self._image.raw_data)

        if gen_mipmap:
            gl.glGenerateMipmap(self._target)

        gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_S, self._wrap_mode.value)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_T, self._wrap_mode.value)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_MIN_FILTER, self._filter_mode.value)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_MAG_FILTER, self._filter_mode.value)

    def _delete(self):
        if self.is_valid():
            gl.glDeleteTextures(1, [self._id])

    @staticmethod
    def _load_image(image_path):
        try:
            img = Image.open(image_path).transpose(Image.FLIP_TOP_BOTTOM)
            num_comps = len(img.getbands())
            tex_format = gl.GL_RED if num_comps == 1 else \
                gl.GL_RGB if num_comps == 3 else gl.GL_RGBA

            return img, tex_format

        except FileNotFoundError:
            import sys
            print(f'Texture failed to load at path: {image_path}', file=sys.stderr)

    def _activate(self):
        self.bind()

    def _deactivate(self):
        self.unbind()

    def bind(self):
        gl.glBindTexture(self._target, self._id)

    def unbind(self):
        gl.glBindTexture(self._target, 0)

    @property
    def wrap_mode(self):
        return self._wrap_mode

    @property
    def filter_mode(self):
        return self._filter_mode

    @property
    def name(self):
        return self._name_path



# MacOS fix
try:
    import OpenGL

    try:
        import OpenGL.GL  # this fails in <=2020 versions of Python on OS X 11.x

    except ImportError:
        print('Patching OpenGL for macOS')
        from ctypes import util

        orig_util_find_library = util.find_library


        def new_util_find_library(name):
            res = orig_util_find_library(name)
            if res:
                return res
            return '/System/Library/Frameworks/' + name + '.framework/' + name


        util.find_library = new_util_find_library
        import OpenGL.GL

except ImportError:
    print('OpenGL import error')
    pass


from OpenGL.GL import *

from OpenGL.GL.shaders import compileShader


def current_context_version():
    return tuple(map(int, glGetString(GL_VERSION).decode('utf-8')[:3].split('.')))



from PIL import Image as PILImage


class Image:
    def __init__(self, pil_image):
        self._pil_image = pil_image
        self._num_channels = len(pil_image.getbands())

    @staticmethod
    def open(file_path):
        img = PILImage.open(file_path).convert('RGBA').transpose(PILImage.FLIP_TOP_BOTTOM)
        return Image(img)

    @property
    def num_channels(self):
        return self._num_channels

    @property
    def raw_data(self):
        return self._pil_image.tobytes()

    @property
    def pil_image(self):
        return self._pil_image

    @property
    def width(self):
        return self._pil_image.width

    @property
    def height(self):
        return self._pil_image.height



import numpy as np
import collections.abc
import enum
import hashlib
from dataclasses import dataclass
from typing import List
from .. import gl
from ..assets.manager import default_asset_mgr, default_name_mgr
from ..assets.resolver import default_resolver
from ..graphics.array import VertexArrayObject
from ..graphics.buffer import VertexBuffer, IndexBuffer
from ..graphics.material import Material
from ..graphics.vertex_layout import VertexLayout, VertexAttrib, VertexAttribDescriptor, VertexAttribFormat
from ..math import Mat4, Vec3
from ..misc import PaletteDefault, Color


# todo: face normals
# todo: sub-meshes and render records


@dataclass(frozen=True)
class Face:
    __slots__ = ('vs', 'vts', 'vns')
    vs: [int]  # vertex indices
    vts: [int]  # vertex texture indices
    vns: [int]  # vertex normal indices

    @property
    def is_triangle(self):
        return len(self.vs) == 3

    @property
    def vertex_count(self):
        return len(self.vs)


@dataclass(frozen=False)
class VertexAttribData:
    """
    Stores the data of a vertex attribute.
    """
    __slots__ = ('count', 'data')
    count: int
    data: np.ndarray

    @property
    def is_empty(self):
        return self.count == 0 and (not self.data.any())

    def __getitem__(self, i):
        return self.data[i]


class MeshTopology(enum.Enum):
    Triangles = gl.GL_TRIANGLES
    Lines = gl.GL_LINES
    LineStrip = gl.GL_LINE_STRIP
    Points = gl.GL_POINTS


@dataclass(frozen=True)
class MtlRenderRecord:
    """
    A render record for a material.
    """
    __slots__ = ['mtl_idx', 'vao_idx', 'vbo_idx', 'vertex_count', 'pipeline_idx']
    mtl_idx: int
    vao_idx: int
    vbo_idx: int
    vertex_count: int
    pipeline_idx: int


class SubMesh:
    """
    Each sub-mesh corresponds to a Material. A sub-mesh consists of a list of triangles,
    which refer to a set of vertices(referenced by index). Vertices can be shared between
    multiple sub-meshes.

    Note: this is a structure maintained inside `Mesh` class, and should not be used directly.
    """

    __slots__ = ['name', 'faces', 'vertex_count', 'topology', 'vertex_layout',
                 'normal_map_enabled', 'bump_map_enabled', 'parallax_map_enabled']

    def __init__(self,
                 faces=None,
                 name='',
                 topology=MeshTopology.Triangles,
                 vertex_layout=VertexLayout.default(),
                 enable_normal_map=False,
                 enable_bump_map=False,
                 enable_parallax_map=False):
        self.name: str = name
        self.faces: [] = [] if faces is None else faces  # index of faces
        self.vertex_count: int = len(faces) * 3 if faces is not None else 0
        self.topology: MeshTopology = topology
        self.vertex_layout: VertexLayout = vertex_layout
        self.normal_map_enabled: bool = enable_normal_map
        self.bump_map_enabled: bool = enable_bump_map
        self.parallax_map_enabled: bool = enable_parallax_map

    @staticmethod
    def from_sub_meshes(sub_meshes, name=''):
        triangles = []
        for sub_mesh in sub_meshes:
            triangles += sub_mesh.faces
        return SubMesh(faces=triangles, name=name)


def _load_pipeline_uniforms(pipeline, excluded=('camera', 'shadow_map'), **kwargs):
    for key, value in kwargs.items():
        if key not in excluded:
            pipeline[key] = value


class Mesh:
    """
    Representation of a generic geometry including vertex positions, face indices, normals, colors, uvs,
    and custom attributes within buffers.

    Mesh contains vertices and multiple triangle arrays. The triangle arrays are indices into the vertex
    arrays; three indices for each triangle.

    For each vertex there can be a normal, eight texture coordinates, color, tangent, and bone weight.
    These are optional and can be removed. All vertex information is stored in separate arrays of the
    same size, so if your mesh has 10 vertices, you would also have 10-size arrays for normals and other
    attributes.

    The mesh face data (triangles) is made of three vertex indices for each triangle.
    """
    __slots__ = ['_vertex_layout', '_index_count', '_verts', '_indices', '_faces',
                 '_faces_triangulation', '_triangle_count', '_triangles', '_materials',
                 '_render_records', '_name', '_sub_meshes_raw', '_sub_meshes', '_sub_mesh_count',
                 '_pipelines', '_init_transform', '_transform', '_shading_enabled',
                 '_normal_map_enabled', '_bump_map_enabled', '_parallax_map_enabled', '_vertex_buffers',
                 '_index_buffers', '_vertex_array_objects', '_cast_shadow', '_material_enabled', '_texture_enabled'
                 ]
    DEFAULT_COLOR = PaletteDefault.BrownB.as_color()

    def __init__(self, name, filepath=None, resolver=default_resolver, **kwargs):
        """
        Creation and initialisation of a Mesh. A mesh can be created by:

        1. reading a OBJ file
        2. providing vertices and its attributes

        Materials(including textures) are specified by specifying the sub-meshes.

        Note:
            If `filepath` is specified, this function will skip arguments inside `kwargs`.

        Args:
            filepath (str):
                Specifies the path of the model file to be loaded.

            **kwargs:
                vertices: (array like):
                    Specifies the position of vertices.

                colors: (Sequence[Color]):
                    Specifies the color of vertices.

                normals: (array like):
                    Specifies the normal of vertices.

                faces: (array like):
                    Specifies the faces of the mesh. Each face is a list consisting of
                    indices of vertices, texture coordinates, and normals. For example,
                    a face can be [(v0, v1, v2), (vt0, vt1, vt2), (vn0, vn1, vn2)], where
                    v0, v1, v2 are indices of vertices, vt0, vt1, vt2 are indices of texture
                    coordinates, and vn0, vn1, vn2 are indices of normals.
                    A list of faces is represented as
                    [
                        [(v0, v1, v2), (vt0, vt1, vt2), (vn0, vn1, vn2)],
                        [(v3, v4, v5, v6), (vt3, vt4, vt5, vt6), (vn3, vn4, vn5, vt6)],
                    ]
                    Faces don't have to be triangles, they can be quads or other polygons,
                    but they have to be convex. Behind the scene, the faces will be triangulated.

                uvs: (array like):
                    Specifies the uv coordinates of the vertices.

                init_transform (Mat4):
                    Specifies the initial transform of the mesh. Can also be specified after the mesh is created.

                transform (Mat4):
                    Specifies the transform of the mesh. Can also be specified after the mesh is created.

                shading_enabled (bool):
                    Specifies whether the shading is enabled (using Phong model).

                material_enabled (bool):
                    Specifies whether the material is enabled. If enabled, the mesh will be rendered with the
                    material loaded from the model file or the default material if no material is loaded. In
                    case the material is not enabled, the mesh will be rendered with vertex colors.

                texture_enabled (bool):
                    Specifies whether the diffuse map will be used.

                cast_shadow (bool):
                    Specifies whether the object will generate shadow.

                vertex_shader (str):
                    Specifies the vertex shader going to be used for rendering. In case of non specified,
                    use framework's default shader (declared inside of Window).

                pixel_shader (str):
                    Specifies the pixel shader going to be used for rendering. In case of non specified,
                    use framework's default shader (declared inside of Window).
        """
        assert name is not None, 'Mesh name cannot be None'
        self._vertex_layout = VertexLayout.default()
        self._index_count = 0
        self._verts = {
            'position': VertexAttribData(0, np.array([], dtype=np.float32)),
            'color': VertexAttribData(0, np.array([], dtype=np.float32)),
            'normal': VertexAttribData(0, np.array([], dtype=np.float32)),
            'texcoord': VertexAttribData(0, np.array([], dtype=np.float32)),
            'tangent': np.array([], dtype=np.float32),
            # stores the indices of triangles that each vertex belongs to; used for calculating tangent and bi-tangent.
            'topology': [[0]],
        }
        self._indices = np.array([], dtype=np.uint32)  # indices for draw whole mesh
        # Face data consists of vertex indices, texture indices, and normal indices.
        # [[(vertices indices), (vertex uvs indices), (vertex normals indices)], ...]
        # Used only when the mesh is created by providing vertices and its attributes.
        self._faces: List[Face] = []

        # List keeping track of triangulated faces in case the mesh is created by providing vertices and their
        # attributes. Stores the first index of triangles of a face after the triangulation in the self._triangles
        # array. Take example of a list of two faces, first face is a quad and second face is a triangle. This list
        # will be [0, 3] after triangulation, where 0 is the index of the first triangle of triangulation of the
        # first face, and 3 is the index of the first triangle of triangulation of the second face.
        self._faces_triangulation = []
        self._triangle_count = 0
        self._triangles = []

        self._materials = [Material("bk7084_default_material")]
        self._render_records = {}  # records the rendering information, (sub_mesh_index: mtl_render_record)
        self._sub_meshes = []
        self._sub_meshes_raw = []  # stores the unprocessed submesh inputs, for easy merging of meshes
        self._sub_mesh_count = 0
        self._pipelines = [
            default_asset_mgr.get_or_create_pipeline('default_pipeline',
                                                     vertex_shader='shaders/default.vert',
                                                     pixel_shader='shaders/default.frag')
        ]
        # initial transformation applied to the mesh
        self._init_transform: Mat4 = kwargs.get('init_transform', Mat4.identity())
        # transformation applied to the mesh
        self._transform: Mat4 = Mat4.identity()

        self._shading_enabled = kwargs.get('shading_enabled', True)
        self._material_enabled = kwargs.get('material_enabled', True)  # by default, use material if available
        self._texture_enabled = kwargs.get('texture_enabled', False)  # by default, texture is not enabled
        self._normal_map_enabled = False
        self._bump_map_enabled = False
        self._parallax_map_enabled = False

        self._vertex_buffers: [VertexBuffer] = []
        self._index_buffers: [IndexBuffer] = []
        self._vertex_array_objects: [VertexArrayObject] = []

        self._cast_shadow = kwargs.get('cast_shadow', True)

        colors = kwargs.get('colors', [Mesh.DEFAULT_COLOR])
        if filepath:  # load from file
            self._name = f"{name}@{default_name_mgr.get_name(hashlib.md5(filepath.encode('utf-8')).hexdigest())}"
            self._load_from_file(filepath, colors, resolver)
        else:  # load from vertices, uvs, normals, faces
            self._name = f"{name}@{default_name_mgr.get_name('customized')}"
            self._verts['position'].data = np.asarray(kwargs['vertices'], dtype=np.float32).reshape(
                (-1, 3)) if 'vertices' in kwargs else None
            self._verts['position'].count = len(self._verts['position'].data)
            self._verts['texcoord'].data = np.asarray(kwargs['uvs'], dtype=np.float32).reshape(
                (-1, 2)) if 'uvs' in kwargs else None
            self._verts['texcoord'].count = len(self._verts['texcoord'].data)
            self._verts['normal'].data = np.asarray(kwargs['normals'], dtype=np.float32).reshape(
                (-1, 3)) if 'normals' in kwargs else None
            self._verts['normal'].count = len(self._verts['normal'].data)
            _faces = kwargs.get('faces', None)

            # Check if the necessary data is provided for creating the mesh.
            if not (self._verts['position'] is not None
                    and self._verts['texcoord'] is not None
                    and self._verts['normal'] is not None
                    and self._faces is not None):
                missing_attributes = [name for attrib, name in ((self._verts['position'], 'vertices'),
                                                                (self._verts['texcoord'], 'uvs'),
                                                                (self._verts['normal'], 'normals'),
                                                                (_faces, 'faces')) if attrib is None]
                raise ValueError(f"Mesh creation - missing vertex attributes: {missing_attributes}.")

            # Check if the input faces are valid.
            for face in _faces:
                if len(face) != 3:
                    raise ValueError(f"Mesh creation - a face must have vertices indices, "
                                     f"texture coordinates indices, and normal indices: {face}.")
                if len(face[0]) < 3:
                    raise ValueError(f"Mesh creation - a face must have at least 3 vertices: {face}.")
                self._faces.append(Face(*face))
            self._verts['color'] = Mesh._process_color(self._verts['position'].count, colors)
            self._verts['topology'] = [[] for _ in range(self._verts['position'].count)]

            # Triangulate faces into triangles and store the indices of triangles of each face.
            _triangles = []
            self._triangle_count = 0
            for f in self._faces:
                if f.is_triangle:
                    self._faces_triangulation.append(len(_triangles))
                    _triangles.append(
                        (
                            (f.vs[0], f.vs[1], f.vs[2]),
                            (f.vts[0], f.vts[1], f.vts[2]),
                            (f.vns[0], f.vns[1], f.vns[2]),
                        )
                    )
                    self._triangle_count += 1
                    for p in f.vs:
                        self._verts['topology'][p].append(self._triangle_count - 1)
                else:
                    self._faces_triangulation.append(len(_triangles))
                    triangulated = []
                    for i in range(0, f.vertex_count - 2):
                        triangulated.append(
                            (
                                (f.vs[0], *f.vs[i + 1: i + 3]),  # vertices
                                (f.vts[0], *f.vts[i + 1: i + 3]),  # texture coordinates
                                (f.vns[0], *f.vns[i + 1: i + 3])  # normals
                            )
                        )
                        self._triangle_count += 1
                        for p in triangulated[-1][0]:
                            self._verts['topology'][p].append(self._triangle_count - 1)
                    _triangles.extend(triangulated)
            self._triangles = np.asarray(_triangles, dtype=np.uint32).reshape((-1, 3, 3))
            self._indices = np.asarray(self._triangles, dtype=np.uint32).ravel()
            self._index_count = len(self._indices)
            if not self._verts['texcoord'].is_empty:
                self._verts['tangents'] = compute_all_tangents(self._verts['position'], self._verts['texcoord'],
                                                               self._triangles, self._verts['topology'])
            else:
                self._verts['tangents'] = np.zeros((self._verts['position'].count, 3), dtype=np.float32)
            self._sub_mesh_count = 1
            sub_mesh = SubMesh(
                name=f"{self._name}_sub_{self._materials[0].name}",
                topology=MeshTopology.Triangles,
                vertex_layout=self._vertex_layout,
                faces=self._faces
            )
            # create rendering record
            v_positions = []
            v_normals = []
            v_uvs = []
            v_colors = []
            v_tangents = []

            for tri in self._triangles:
                for i, content in enumerate(tri):
                    if i == 0:
                        for v_i in content:
                            v_positions.append(self._verts['position'][v_i])
                            v_colors.append(self._verts['color'][v_i])
                            v_tangents.append(self._verts['tangents'][v_i])
                    if i == 1:
                        for vt_i in content:
                            v_uvs.append(self._verts['texcoord'][vt_i])
                    if i == 2:
                        for vn_i in content:
                            v_normals.append(self._verts['normal'][vn_i])

            v_vertices = np.array(
                [z for x in zip(v_positions, v_colors, v_uvs, v_normals, v_tangents) for y in x for z in y],
                dtype=np.float32).ravel()

            vao = VertexArrayObject()
            vbo = VertexBuffer(len(v_positions), sub_mesh.vertex_layout)
            vbo.set_data(v_vertices)
            vao.bind_vertex_buffer(vbo, [0, 1, 2, 3, 4])

            self._sub_meshes = [sub_mesh]
            # TODO: store sub-meshes pipelines
            self._render_records = {
                0: MtlRenderRecord(0, len(self._vertex_array_objects), len(self._vertex_buffers), len(v_positions), 0)
            }

            self._vertex_array_objects.append(vao)
            self._vertex_buffers.append(vbo)

    @staticmethod
    def _process_color(count, colors):
        output_colors = None
        if isinstance(colors, Color):
            output_colors = np.tile(colors.rgba, (count, 1))
        elif isinstance(colors, collections.abc.Sequence):
            colors_count = len(colors)
            if colors_count == 1:
                output_colors = np.tile([colors[0].rgba], (count, 1))
            elif colors_count >= count:
                output_colors = np.asarray([c.rgba for c in colors[:count]]).reshape((count, 4))
            else:
                num = count - colors_count
                output_colors = np.asarray(
                    [c.rgba for color in [colors, [Mesh.DEFAULT_COLOR] * num] for c in color]).reshape((count, 4))
        return VertexAttribData(count, output_colors)

    @property
    def rendering_info(self):
        info = {}
        for sub_mesh_idx, record in self._render_records.items():
            pipeline = self._pipelines[record.pipeline_idx]
            vao = self._vertex_array_objects[record.vao_idx]
            vertex_count = record.vertex_count
            mtl = self._materials[record.mtl_idx]
            topology = self._sub_meshes[sub_mesh_idx].topology.value
            diffuse_map = mtl.diffuse_map
            transform = self._transform * self._init_transform
            if pipeline.uuid not in info:
                info[pipeline.uuid] = []
            info[pipeline.uuid].append((vao, topology, vertex_count, mtl,
                                        diffuse_map,
                                        self._shading_enabled,
                                        self._material_enabled,
                                        self._texture_enabled,
                                        self._normal_map_enabled,
                                        self._bump_map_enabled,
                                        self._parallax_map_enabled,
                                        transform))
        return info

    @property
    def shading_enabled(self):
        return self._shading_enabled

    @shading_enabled.setter
    def shading_enabled(self, value):
        self._shading_enabled = value

    @property
    def texture_enabled(self):
        return self._texture_enabled

    @texture_enabled.setter
    def texture_enabled(self, value):
        self._texture_enabled = value

    @property
    def normal_map_enabled(self):
        return self._normal_map_enabled

    @normal_map_enabled.setter
    def normal_map_enabled(self, value):
        self._normal_map_enabled = value

    @property
    def bump_map_enabled(self):
        return self._bump_map_enabled

    @bump_map_enabled.setter
    def bump_map_enabled(self, value):
        self._bump_map_enabled = value

    @property
    def parallax_map_enabled(self):
        return self._parallax_map_enabled

    @parallax_map_enabled.setter
    def parallax_map_enabled(self, value):
        self._parallax_map_enabled = value

    @property
    def material_enabled(self):
        return self._material_enabled

    @material_enabled.setter
    def material_enabled(self, value):
        self._material_enabled = value

    @property
    def cast_shadow(self):
        return self._cast_shadow

    @cast_shadow.setter
    def cast_shadow(self, value):
        self._cast_shadow = value

    def _load_from_file(self, filepath, colors, resolver=default_resolver):
        mesh_data = default_asset_mgr.get_or_load_wavefront_obj(filepath, resolver)
        self._verts['position'].data = np.asarray(mesh_data['positions'], dtype=np.float32).reshape((-1, 3))
        self._verts['position'].count = len(mesh_data['positions'])
        self._verts['normal'].data = np.asarray(mesh_data['normals'], dtype=np.float32).reshape((-1, 3))
        self._verts['normal'].count = len(mesh_data['normals'])
        self._faces = mesh_data['faces']
        self._triangles = np.array(self._faces).reshape((-1, 3, 3))
        self._triangle_count = len(self._triangles)
        self._verts['topology'] = [[] for _ in range(0, self._verts['position'].count)]
        for i, tri in enumerate(self._triangles):
            for p in tri[0]:
                self._verts['topology'][p].append(i)

        self._index_count = self._triangle_count * 3
        self._indices = np.asarray([f[0] for f in mesh_data['faces']], dtype=np.uint32).reshape((-1, 3))
        self._verts['texcoord'].count = len(mesh_data['texcoords'])
        self._verts['texcoord'].data = np.asarray(mesh_data['texcoords'], dtype=np.float32).reshape((-1, 2))
        self._verts['color'] = Mesh._process_color(self._verts['position'].count, colors)
        if not self._verts['texcoord'].is_empty:
            self._verts['tangent'] = compute_all_tangents(self._verts['position'], self._verts['texcoord'],
                                                          self._triangles, self._verts['topology'])
        else:
            self._verts['tangent'] = np.zeros((self._verts['position'].count, 3), dtype=np.float32)

        # Loaded model has a name like "{hash(filepath)}_000", to avoid creating multiple times the same materials,
        # we remove the _000 suffix from the name, so that the material is shared between all meshes with the same name.
        for name, mat in mesh_data['materials'].items():
            mtl_name = f"{'_'.join(self._name.split('_')[:-1])}_{name}"
            self._materials.append(
                Material(
                    mtl_name,
                    ka=mat.get('Ka', [0.8, 0.8, 0.8]),
                    kd=mat.get('Kd', [0.8, 0.8, 0.8]),
                    ks=mat.get('Ks', [1.0, 1.0, 1.0]),
                    shininess=mat.get('Ns', 1.0),
                    ior=mat.get('Ni', 1.0),
                    dissolve=mat.get('d', 1.0),
                    illum=mat.get('illum', 0),
                    diffuse_map_path=mat['map_Kd']['path'] if 'map_Kd' in mat else None,
                    bump_map_path=mat['map_bump']['path'] if 'map_bump' in mat else None,
                    normal_map_path=mat['map_norm']['path'] if 'map_norm' in mat else None
                )
            )

        for sub_obj in mesh_data['objects']:
            sub_obj_materials = []
            for mtl in sub_obj['materials']:
                mtl_name = f"{'_'.join(self._name.split('_')[:-1])}_{mtl['name']}"
                f_start, f_end = tuple(mtl['f_range'])
                mtl_idx = self._materials.index([x for x in self._materials if x.name == mtl_name][0])
                if mtl_idx != -1 and mtl_idx < len(self._materials):
                    sub_obj_materials.append((mtl_idx, f_start, f_end))

            # will populate a default color
            color = self._verts['color'][0]
            vertex_attribs = [VertexAttribDescriptor(VertexAttrib.Position, VertexAttribFormat.Float32, 3),
                              VertexAttribDescriptor(VertexAttrib.Color0, VertexAttribFormat.Float32, 4)]

            for attrib in sub_obj['vertex_format']:
                if attrib == 'T':
                    vertex_attribs.append(VertexAttribDescriptor(VertexAttrib.TexCoord0, VertexAttribFormat.Float32, 2))
                if attrib == 'N':
                    vertex_attribs.append(VertexAttribDescriptor(VertexAttrib.Normal, VertexAttribFormat.Float32, 3))

            vertex_attribs.append(VertexAttribDescriptor(VertexAttrib.Tangent, VertexAttribFormat.Float32, 3))

            # pack the data together for rendering
            for mtl in sub_obj_materials:
                mtl_idx, f_start, f_end = mtl
                sub_mesh = SubMesh(
                    name=f"{sub_obj['name']}-{self._materials[mtl[0]].name}",
                    faces=list(range(f_start, f_end))
                )
                sub_mesh.vertex_layout = VertexLayout(*vertex_attribs)
                v_positions = []
                v_normals = []
                v_texcoords = []
                v_tangents = []
                for face in self._faces[f_start: f_end]:
                    i = 0
                    for v_i in face[i]:
                        v_positions.append(self._verts['position'][v_i])
                        v_tangents.append(self._verts['tangent'][v_i])
                    i += 1
                    if sub_mesh.vertex_layout.has(VertexAttrib.TexCoord0):
                        for vt_i in face[i]:
                            v_texcoords.append(self._verts['texcoord'][vt_i])
                        i += 1
                    if sub_mesh.vertex_layout.has(VertexAttrib.Normal):
                        for vn_i in face[i]:
                            v_normals.append(self._verts['normal'][vn_i])

                # create one interleaved buffer
                vertex_count = len(v_positions)
                v_colors = [color] * vertex_count

                if sub_mesh.vertex_layout.has(VertexAttrib.TexCoord0) and sub_mesh.vertex_layout.has(
                        VertexAttrib.Normal):
                    vertices = np.array(
                        [z for x in zip(v_positions, v_colors, v_texcoords, v_normals, v_tangents) for y in x for z in
                         y],
                        dtype=np.float32).ravel()
                elif sub_mesh.vertex_layout.has(VertexAttrib.TexCoord0) and not sub_mesh.vertex_layout.has(
                        VertexAttrib.Normal):
                    vertices = np.array(
                        [z for x in zip(v_positions, v_colors, v_texcoords, v_tangents) for y in x for z in y],
                        dtype=np.float32).ravel()
                elif not sub_mesh.vertex_layout.has(VertexAttrib.TexCoord0) and sub_mesh.vertex_layout.has(
                        VertexAttrib.Normal):
                    vertices = np.array(
                        [z for x in zip(v_positions, v_colors, v_normals, v_tangents) for y in x for z in y],
                        dtype=np.float32).ravel()
                else:
                    vertices = np.array([z for x in zip(v_positions, v_colors, v_tangents) for y in x for z in y],
                                        dtype=np.float32).ravel()

                # bind vertex buffers
                vao = VertexArrayObject()
                vbo = VertexBuffer(vertex_count, sub_mesh.vertex_layout)
                vbo.set_data(vertices)

                vbo_idx = len(self._vertex_buffers)
                self._vertex_buffers.append(vbo)
                vao_idx = len(self._vertex_array_objects)
                self._vertex_array_objects.append(vao)

                if not sub_mesh.vertex_layout.has(VertexAttrib.TexCoord0):
                    vao.bind_vertex_buffer(vbo, [0, 1, 3, 4])
                else:
                    vao.bind_vertex_buffer(vbo, [0, 1, 2, 3, 4])

                self._sub_meshes.append(sub_mesh)

                sub_mesh_index = len(self._sub_meshes) - 1

                self._render_records[sub_mesh_index] = MtlRenderRecord(mtl_idx, vao_idx, vbo_idx, vertex_count,
                                                                       0)
        self._sub_mesh_count = len(self._sub_meshes)

    @property
    def vertex_layout(self) -> VertexLayout:
        return self._vertex_layout

    @vertex_layout.setter
    def vertex_layout(self, layout: VertexLayout):
        self._vertex_layout = layout
        # TODO: update vertex attribute pointer

    @property
    def vertex_attrib_count(self):
        """Returns the number of vertex attributes that the mesh has."""
        return self._vertex_layout.attrib_count

    @property
    def vertex_buffer_count(self):
        """Returns the number of vertex buffers present in the mesh."""
        return len(self._vertex_buffers)

    @property
    def vertices(self):
        """Returns the vertex positions."""
        return self._verts['position'].data

    @vertices.setter
    def vertices(self, positions):
        self._verts['position'].data = np.asarray(positions, dtype=np.float32).ravel()
        self._verts['position'].count = len(self._verts['position'].data) / 3
        print("Vertices: ", self._verts['position'].data)
        print("Vertices count: ", self._verts['position'].count)

    @property
    def texcoords(self):
        """Returns the vertex uvs."""
        return self._verts['texcoord'].data

    @property
    def vertex_count(self):
        """Vertex count of the mesh."""
        return self._verts['position'].count

    @property
    def colors(self):
        """Vertex colors of the mesh."""
        return self._verts['color']

    @colors.setter
    def colors(self, colors):
        """Vertex colors of the mesh."""
        self._verts['color'] = np.asarray(colors, dtype=np.float32).ravel()

    @property
    def normals(self):
        """Vertex normals of the mesh."""
        return self._verts['normal'].data

    @property
    def triangles(self):
        """Triangles of the mesh."""
        return self._triangles

    @property
    def materials(self):
        return self._materials

    @property
    def material_count(self):
        return len(self._materials)

    @property
    def render_records(self):
        return self._render_records

    @property
    def sub_meshes(self):
        return tuple(self._sub_meshes)

    @property
    def sub_mesh_count(self):
        return self._sub_mesh_count

    @property
    def sub_meshes_raw(self):
        return self._sub_meshes_raw

    def update_sub_mesh(self, index, new: SubMesh, texture: str = None, normal_map: str = None,
                        vertex_shader: str = None, pixel_shader: str = None, create: bool = False):
        if len(self.sub_meshes_raw) == 0:
            self._sub_meshes_raw.append((new, texture, normal_map))

        sub_mesh = self._sub_meshes[index]
        sub_mesh.name = new.name
        sub_mesh.vertex_count = new.vertex_count
        sub_mesh.faces = new.faces
        sub_mesh.topology = new.topology
        sub_mesh.normal_map_enabled = new.normal_map_enabled

        tri_list = []
        vertex_triangles = {}
        for f_i in sub_mesh.faces:
            start = self._faces_triangulation[f_i]
            end = len(self._triangles) if f_i >= len(self._faces_triangulation) - 1 else \
                self._faces_triangulation[f_i + 1]
            tri_list.extend(self._triangles[start: end])
            for i, tri in enumerate(self._triangles[start: end]):
                for v_i in tri[0]:
                    if v_i not in vertex_triangles:
                        vertex_triangles[v_i] = []
                    vertex_triangles[v_i].append(start + i)

        tangents = {}
        for v_i in vertex_triangles.keys():
            tangents[v_i] = self._compute_tangent_of_vertex(v_i, vertex_triangles)

        # create new vbo with new data
        v_positions = []
        v_normals = []
        v_uvs = []
        v_colors = []
        v_tangents = []
        for tri in tri_list:
            for i in range(0, 3):
                if i == 0:
                    for v_i in tri[i]:  # positions
                        v_positions.append(self._verts['position'][v_i])
                        v_colors.append(self._verts['color'][v_i])
                        v_tangents.append(tangents[v_i])
                if i == 1:
                    for vt_i in tri[i]:  # uvs
                        v_uvs.append(self._verts['texcoord'][vt_i])
                if i == 2:
                    for vn_i in tri[i]:  # normals
                        v_normals.append(self._verts['normal'][vn_i])

        sub_mesh.vertex_count = len(v_positions)

        vertices = np.array([z for x in zip(v_positions, v_colors, v_uvs, v_normals, v_tangents) for y in x for z in y],
                            dtype=np.float32).ravel()

        mtl_idx = 0

        if create:
            vao = VertexArrayObject()
            vao_idx = len(self._vertex_array_objects)
            self._vertex_array_objects.append(vao)
        else:
            old_record = self._render_records[index]
            self._vertex_buffers[old_record.vbo_idx].delete()
            vao = self._vertex_array_objects[old_record.vao_idx]
            vao_idx = old_record.vao_idx
            mtl_idx = old_record.mtl_idx

        layout = VertexLayout(
            VertexAttribDescriptor(VertexAttrib.Position, VertexAttribFormat.Float32, 3),
            VertexAttribDescriptor(VertexAttrib.Color0, VertexAttribFormat.Float32, 4),
            VertexAttribDescriptor(VertexAttrib.TexCoord0, VertexAttribFormat.Float32, 2),
            VertexAttribDescriptor(VertexAttrib.Normal, VertexAttribFormat.Float32, 3),
            VertexAttribDescriptor(VertexAttrib.Tangent, VertexAttribFormat.Float32, 3)
        )

        vbo = VertexBuffer(sub_mesh.vertex_count, layout)
        vbo.set_data(vertices)

        vbo_idx = len(self._vertex_buffers)
        self._vertex_buffers.append(vbo)

        vao.bind_vertex_buffer(vbo, [0, 1, 2, 3, 4])

        if texture is not None or normal_map is not None:
            mtl_idx = len(self._materials)
            self._materials.append(
                Material(
                    f'sub_mesh_material_{hashlib.md5(texture.encode("utf-8")).hexdigest()}',
                    diffuse_map_path=texture,
                    normal_map_path=normal_map,
                )
            )
            self._texture_enabled = True

        pipeline = 0
        if vertex_shader is not None or pixel_shader is not None:
            self._pipelines.append(
                default_asset_mgr.get_or_create_pipeline(
                    f'{id(self)}_sub_mesh_{id(sub_mesh)}_pipeline',
                    vertex_shader, pixel_shader
                )
            )
            pipeline = len(self._pipelines) - 1

        self._render_records[index] = MtlRenderRecord(mtl_idx,
                                                      vao_idx,
                                                      vbo_idx,
                                                      sub_mesh.vertex_count, pipeline)

    def append_sub_mesh(self, sub_mesh: SubMesh, texture: str = '', normal_map: str = None, vertex_shader: str = None,
                        pixel_shader: str = None):
        self._sub_meshes_raw.append((sub_mesh, texture, normal_map))
        self._sub_meshes.append(SubMesh())
        self.update_sub_mesh(len(self._sub_meshes) - 1, sub_mesh, texture, normal_map, vertex_shader, pixel_shader,
                             create=True)

    def _compute_tangent_of_vertex(self, vertex_id, vertex_triangles):
        """
        Compute tangent from Tangent space for given vertex (over whole mesh).
        """
        tangent = np.array([0.0, 0.0, 0.0], dtype=np.float32)

        if self._verts['texcoord'].is_empty:
            return np.array([0.0, 0.0, 0.0], dtype=np.float32)
        for tri_idx in vertex_triangles[vertex_id]:
            vertices = [self._verts['position'][i] for i in self._triangles[tri_idx][0]]
            uvs = [self._verts['texcoord'][i] for i in self._triangles[tri_idx][1]]
            # edges (delta vertex positions)
            delta_pos = [vertices[i] - vertices[0] for i in (1, 2)]
            # delta uvs
            delta_uv = [uvs[i] - uvs[0] for i in (1, 2)]
            det = delta_uv[0][0] * delta_uv[1][1] - delta_uv[0][1] * delta_uv[1][0]
            if det != 0.0:
                tangent += (delta_pos[0] * delta_uv[1][1] - delta_pos[1] * delta_uv[0][1]) / det
        return tangent / np.sqrt(np.sum(tangent ** 2))

    def apply_transform(self, matrix: Mat4):
        self._transform = matrix * self._transform
        return self

    def then(self, matrix: Mat4):
        self._transform = matrix * self._transform
        return self

    def reset_transform(self):
        self._transform = Mat4.identity()

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value: Mat4):
        self._transform = value

    @property
    def init_transform(self):
        return self._init_transform

    @init_transform.setter
    def init_transform(self, value: Mat4):
        self._init_transform = value

    @property
    def vertex_array_objects(self):
        return self._vertex_array_objects

    def compute_energy(self, shader, transform, light, viewport_size, depth_map):
        model_mat = transform * self._transform * self._init_transform
        if len(self._sub_meshes) > 0:
            with shader:
                shader['model_mat'] = model_mat
                shader['depth_map'] = 0
                shader['light_mat'] = light.matrix
                shader['light_pos'] = light.position
                shader['light_view_mat'] = light.view_matrix
                shader['resolution'] = viewport_size
                for idx, record in self._render_records.items():
                    sub_mesh = self._sub_meshes[idx]
                    vao = self._vertex_array_objects[record.vao_idx]
                    shader.active_texture_unit(0)
                    with depth_map:
                        with vao:
                            gl.glDrawArrays(sub_mesh.topology.value, 0, record.vertex_count)

    def draw(self, matrix=Mat4.identity(), shader=None, shader_params=None, **kwargs):
        if self._sub_mesh_count > 0:
            for idx, record in self._render_records.items():
                sub_mesh = self._sub_meshes[idx]
                mtl = self._materials[record.mtl_idx]
                mat = matrix * self._transform * self._init_transform
                vao = self._vertex_array_objects[record.vao_idx]
                pipeline = self._pipelines[record.pipeline_idx] if shader is None else shader

                camera_enabled = kwargs.get('camera_enabled', True)
                from .. import app
                camera = kwargs.get('camera', app.current_window().camera)

                with pipeline:
                    for key, value in kwargs.items():
                        if key != 'camera' and key != 'shadow_map':
                            pipeline[key] = value

                    if camera_enabled:
                        pipeline['view_mat'] = camera.view_matrix
                        pipeline['proj_mat'] = camera.projection_matrix

                    pipeline['model_mat'] = mat
                    pipeline['shading_enabled'] = self._shading_enabled
                    pipeline['in_light_pos'] = kwargs.get('in_light_pos', Vec3(600.0, 600.0, 600.0))
                    pipeline['is_directional'] = kwargs.get('is_directional', False)
                    pipeline['in_light_dir'] = kwargs.get('in_light_dir', Vec3(1.0, 1.0, 1.0).normalised)
                    pipeline['light_color'] = kwargs.get('light_color', Vec3(0.8, 0.8, 0.8))

                    pipeline['mtl.diffuse'] = mtl.diffuse
                    pipeline['mtl.ambient'] = mtl.ambient
                    pipeline['mtl.specular'] = mtl.specular
                    pipeline['mtl.shininess'] = mtl.shininess
                    pipeline['mtl.enabled'] = self._material_enabled
                    pipeline['time'] = app.current_window().elapsed_time
                    pipeline['mtl.use_diffuse_map'] = self._texture_enabled

                    if self._sub_mesh_count > 1:
                        pipeline['mtl.use_normal_map'] = sub_mesh.normal_map_enabled
                    else:
                        pipeline['mtl.use_normal_map'] = self._normal_map_enabled

                    pipeline['mtl.use_bump_map'] = self._bump_map_enabled
                    pipeline['mtl.use_parallax_map'] = self._parallax_map_enabled
                    pipeline['mtl.diffuse_map'] = 0
                    pipeline['mtl.bump_map'] = 1
                    pipeline['mtl.normal_map'] = 2

                    shadow_map_enabled = kwargs.get('shadow_map_enabled', False)
                    depth_map = kwargs.get('shadow_map', None)

                    if shader_params is not None and isinstance(shader_params, dict):
                        for key, value in shader_params.items():
                            pipeline[key] = value

                    pipeline.active_texture_unit(0)
                    with mtl.diffuse_map:
                        pipeline.active_texture_unit(1)
                        with mtl.bump_map:
                            pipeline.active_texture_unit(2)
                            with mtl.normal_map:
                                if shadow_map_enabled and depth_map is not None:
                                    pipeline.active_texture_unit(3)
                                    pipeline['shadow_map'] = 3
                                    with depth_map:
                                        with vao:
                                            gl.glDrawArrays(sub_mesh.topology.value, 0, record.vertex_count)
                                else:
                                    with vao:
                                        gl.glDrawArrays(sub_mesh.topology.value, 0, record.vertex_count)


def compute_all_tangents(positions: VertexAttribData,
                         texcoords: VertexAttribData,
                         triangles: np.ndarray,
                         topology: List[List[int]]) -> np.ndarray:
    """
    Compute the tangent vector for all vertices in the mesh.
    Note that only tangent is needed, bi-tangent can be computed in shader by cross product: N*T

        e = edge in triangle space
        s = edge in texture space

        we want to find the transformation from texture space to triangle space

        [T_x B_x]   [s1_x s2_x]   [e1_x e2_x]
        [T_y B_y] * [s1_y s2_y] = [e1_y e2_y]
        [T_z B_z]                 [e1_z e2_z]

        by inverting the 2x2 matrix

        [T_x B_x]   [e1_x e2_x]   [ s2_y -s2_x]
        [T_y B_y] = [e1_y e2_y] * [-s1_y  s1_x] * 1/det
        [T_z B_z]   [e1_z e2_z]

        where:
        det = s1_x*s2_y - s1_y*s2_x

        now we can solve for T:

        T_x = e1_x * s2_y - e2_x * s1_y
        T_y = e1_y * s2_y - e2_y * s1_y
        T_z = e1_z * s2_y - e2_z * s1_y

        or, in short:

        T = e1 * s2_y - e2 * s1_y

    Args:
        positions (np.ndarray): The positions of all vertices in the mesh.
        texcoords (np.ndarray): The texture coordinates of all vertices in the mesh.
        triangles (np.ndarray): The triangles in the mesh, it is a 3D array with shape (n, 3, 3). First dimension is the
                                triangle index, second dimension is the attributes of the vertex in the triangle,
                                third dimension is the attribute values. For example, triangles[0, 0] is the vertices
                                of the first triangle, triangles[0, 1] is the texture coordinates of the first triangle
                                and triangles[0, 2] is the normals of the first triangle.
        topology (List[List[int]]): The topology information of the mesh. This is list containing indices of all
                                    triangles adjacent to a vertex.
    Returns:
        A list of tangent vectors for all vertices in the mesh.
    """
    count = positions.count
    tangents = np.zeros((count, 3), dtype=np.float32)
    for i in range(count):
        tangent = np.array([0.0, 0.0, 0.0], dtype=np.float32)
        for adj_tri in topology[i]:
            verts_pos = triangles[adj_tri][0]
            verts_tex = triangles[adj_tri][1]
            vs = [positions[verts_pos[0]], positions[verts_pos[1]], positions[verts_pos[2]]]
            uvs = [texcoords[verts_tex[0]], texcoords[verts_tex[1]], texcoords[verts_tex[2]]]
            delta_pos = [vs[1] - vs[0], vs[2] - vs[0]]
            delta_uv = [uvs[1] - uvs[0], uvs[2] - uvs[0]]
            det = delta_uv[0][0] * delta_uv[1][1] - delta_uv[0][1] * delta_uv[1][0]
            if det != 0.0:
                tangent += (delta_pos[0] * delta_uv[1][1] - delta_pos[1] * delta_uv[0][1]) / det
        norm = np.sqrt(np.sum(tangent ** 2))
        if norm != 0.0:
            tangents[i] = tangent / norm
    return tangents



from typing import Sequence

import numpy as np

from .shape import Shape
from ..graphics.util import DrawingMode
from ..math import Vec3
from ..misc import Color, PaletteDefault


class Box(Shape):
    def __init__(self, width: float = 1.0, height: float = 1.0, depth: float = 1.0,
                 colors: Sequence[Color] = (PaletteDefault.YellowB.as_color(),)):
        super().__init__(8, colors)
        self._width = width
        self._height = height
        self._depth = depth

        self._vertices, self._indices = Box.gen_vertices_and_indices(width, height, depth)

    @staticmethod
    def gen_vertices_and_indices(w, h, d):
        vertices = []

        start = Vec3(-w * 0.5, -h * 0.5, -d * 0.5)

        for ix in range(0, 2):
            for iz in range(0, 2):
                for iy in range(0, 2):
                    vertices.append(start + Vec3(ix * w, iy * h, iz * d))

        indices = np.array([0, 1, 2, 2, 1, 3,
                            4, 5, 0, 0, 5, 1,
                            6, 7, 4, 4, 7, 5,
                            2, 3, 6, 6, 3, 7,
                            5, 7, 1, 1, 7, 3,
                            6, 4, 2, 2, 4, 0], dtype=np.uint32)

        return np.asarray(vertices, dtype=np.float32).ravel(), indices

    @property
    def vertices(self) -> np.ndarray:
        return self._vertices

    @property
    def vertex_count(self) -> int:
        return len(self._vertices) // 3

    @property
    def drawing_mode(self) -> DrawingMode:
        return DrawingMode.Triangles

    @property
    def indices(self):
        return self._indices

    @property
    def index_count(self):
        return len(self._indices)



from .shape import Shape
from .ray import Ray
from .triangle import Triangle
from .line import Line
from .box import Box
from .grid import Grid, AxisAlignment

# TODO:
#    Box
#    Circle
#    Cone
#    Cylinder
#    Dodecahedron
#    Icosahedron
#    Plane
#    Polyhedron
#    Ring
#    Tetrahedron
#    Torus
#    TorusKnot



import inspect
import logging


class EventDispatcher:
    """Generic event dispatcher that listens and dispatches events.

    For each type of event, there can only be multiple handlers attached to it.

    @dispatcher.event
    def event_name():
        pass

    @dispatcher.event('event_name')
    def random_name():
        pass
    """
    _event_types = []

    def __init__(self):
        self._event_listeners = {}

    @classmethod
    def register_event_type(cls, name):
        """Register a certain type of event within the dispatcher.

        Args:
            name (str): Name of the event to register.

        Returns:
            None
        """
        cls._event_types.append(name)

    def attach_listener(self, event, listener):
        """Attach an event listener for an event type."""
        if event not in self._event_types:
            print(f'Warn: {event} not yet registered!')
        else:
            if event not in self._event_listeners:
                self._event_listeners[event] = []
            self._event_listeners[event].append(listener)

    def attach_listeners(self, obj):
        if not isinstance(obj, (list, tuple)):
            for event in self._event_types:
                if hasattr(obj, event):
                    self.attach_listener(event, getattr(obj, event))

    def detach_listener(self, event, listener):
        """Remove event listener."""
        if listener in self._event_listeners[event]:
            self._event_listeners[event].remove(listener)

    def detach_listeners(self, obj):
        if not isinstance(obj, (list, tuple)):
            for event in self._event_types:
                if hasattr(obj, event):
                    self.detach_listener(event, getattr(obj, event))

    def dispatch(self, event, *args, **kwargs):
        """Dispatch an event to attached handlers."""
        logging_enabled = kwargs.get('logging_enabled', False)
        reversed_exec_order = kwargs.get('reversed_exec_order', False)
        # Search in instance
        if hasattr(self, event):
            try:
                getattr(self, event)(*args)
            except TypeError:
                self._raise_dispatch_exception(event, args, getattr(self, event))

        # Search in listeners
        listeners = self._event_listeners.get(event, None)
        if listeners:
            try:
                if reversed_exec_order:
                    for listener in listeners:
                        if logging_enabled:
                            logging.info(f'dispatch <{event}> to <{listener}>')
                        listener(*args)
                else:
                    for listener in listeners[::-1]:
                        if logging_enabled:
                            logging.info(f'dispatch <{event}> to <{listener}>')
                        listener(*args)
            except TypeError:
                self._raise_dispatch_exception(event, args, getattr(self, event))

    @staticmethod
    def _raise_dispatch_exception(event, args, listener):
        n_args = len(args)
        spec = inspect.getfullargspec(listener)
        n_listener_args = len(spec.args)

        if inspect.ismethod(listener) and listener.__self__:
            n_listener_args -= 1

        # Allow *args varargs to over specify arguments
        if spec.varargs:
            n_listener_args = max(n_listener_args, n_args)

        # Allow default values to over specify arguments
        if n_listener_args > n_args >= n_listener_args - len(spec.defaults) and spec.defaults:
            n_listener_args = n_args

        if n_listener_args != n_args:
            if inspect.isfunction(listener) or inspect.ismethod(listener):
                desc = f'{listener.__name__} at {listener.__code__.co_filename}:{listener.__code__.co_firstlineno}'
            else:
                desc = repr(listener)

            raise TypeError(f'{event} event was dispatched with {n_args} arguments, but '
                            f'listener {desc} has an incompatible function signature')
        else:
            raise

    def event(self, *args):
        """Function decorator for an event listener/handler.
        """
        if len(args) == 0:  # @window.event()
            def decorator(fn):
                fn_name = fn.__name__
                self.attach_listener(fn_name, fn)
                return fn

            return decorator

        elif inspect.isroutine(args[0]):  # @window.event
            fn = args[0]
            fn_name = fn.__name__
            self.attach_listener(fn_name, fn)
            return args[0]

        elif type(args[0]) in (str,):  # @window.event('on_resize')
            fn_name = args[0]

            def decorator(fn):
                self.attach_listener(fn_name, fn)

            return decorator



from .. import gl


def draw(obj):
    from ..geometry.triangle import Triangle
    from ..geometry.ray import Ray

    if isinstance(obj, Triangle):
        gl.glColor4f(*obj.color)
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, obj.vertices)
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, 3)
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)

    if isinstance(obj, Ray):
        gl.glColor4f(*obj.color)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(obj.origin.x, obj.origin.y, obj.origin.z)
        gl.glVertex3f(obj.origin.x + obj.direction.x, obj.origin.y + obj.direction.y, obj.origin.z + obj.direction.z)
        gl.glEnd()



from .light import Light
from ...math import Vec3, Mat4
from ...misc import Color, PaletteSvg


class DirectionalLight(Light):
    def __init__(self, position: Vec3 = Vec3(8.0, 8.0, 8.0), direction: Vec3 = Vec3(-1.0, -1.0, -1.0),
                 color: Color = PaletteSvg.White.as_color(), sm_width=20.0, sm_height=20.0, **kwargs):
        """

        Args:
            direction:
            color:
            sm_width (int): shadow map width
            sm_height (int): shadow map height
            **kwargs:
                sm_near (float): shadow map near plane
                sm_far (float): shadow map far plane
        """
        super(DirectionalLight, self).__init__(position, color, sm_width=sm_width, sm_height=sm_height, **kwargs)
        self._direction = direction.normalised
        self._is_dirty = True

        self._update()

    def _update(self):
        up = Vec3.unit_y()
        if self._direction == -Vec3.unit_y():
            up = Vec3.unit_x()
        elif self._direction == -Vec3.unit_y():
            up = Vec3.unit_z()
        self._view_mat = Mat4.look_at_gl(Vec3(self._position),
                                         (self._position + self._direction).normalised,
                                         up)
        self._proj_mat = Mat4.orthographic_gl(-self._sm_width / 2.0, self._sm_width / 2.0,
                                              -self._sm_height / 2.0, self._sm_height / 2.0,
                                              self._sm_near, self._sm_far)
        self._is_dirty = False

    @property
    def matrix(self):
        if self._is_dirty:
            self._update()
        return self._proj_mat * self._view_mat

    @property
    def view_matrix(self):
        if self._is_dirty:
            self._update()
        return self._view_mat

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        self._direction = Vec3(value).normalised
        self._position = self._direction / 0.577 * -8.0
        self._is_dirty = True

    @property
    def is_directional(self):
        return True

    @property
    def is_perspective(self):
        return False



# TODO: supports other kind of shaders
# TODO: parse uniforms and generate its getter and setter
import logging
import platform
from collections.abc import Sequence

import OpenGL.GL as gl
import numpy as np

from .shader import Shader
from .util import DrawingMode, GpuObject, BindSemanticObject
from .variable import gl_get_uniform, glsl_types, gl_set_uniform, gl_type_info


def check_validity(program_id) -> bool:
    """Validates the shader program."""
    gl.glValidateProgram(program_id)
    status = gl.glGetProgramiv(program_id, gl.GL_VALIDATE_STATUS)
    return False if status == gl.GL_FALSE else True


def check_link_status(program_id) -> bool:
    """Check link status of shader program."""
    link_status = gl.glGetProgramiv(program_id, gl.GL_LINK_STATUS)
    return False if link_status == gl.GL_FALSE else True


class ShaderProgram(GpuObject, BindSemanticObject):
    __slots__ = ('_id', '_shaders', '_uniforms', '_attributes', '_is_active', '_current_activated_textured_unit')

    def __init__(self, *shaders: Shader):
        """Creates an OpenGL ShaderProgram from multiple shaders.

        Args:
            *shaders:
        """
        super().__init__(None, -1)
        assert len(shaders) > 1, "At least two Shader objects(vertex shader, fragment shader) are required."
        self._id = self._create_and_link(shaders)
        self._is_active = False

        self._buffers = None
        self._shaders = list(shaders)

        self._uniforms = {k: v for k, v in
                          [(x[0], (glsl_types[x[1]], gl.glGetUniformLocation(self._id, x[0]))) for s in self._shaders
                           for x in s.uniforms]}

        self._attributes = {}
        self._current_activated_textured_unit = -1

        # Allows get/set shader uniform as if it is an attribute of the class.
        for uniform in self._uniforms.keys():
            setattr(self.__class__, uniform, self.__create_uniform_prop(uniform))

    def __create_uniform_prop(self, name):
        @property
        def prop(self):
            return self.get_uniform(name)

        @prop.setter
        def prop(self, value):
            self.set_uniform(name, value)

        return prop

    def __getitem__(self, uniform: str):
        return self.get_uniform(uniform)

    def __setitem__(self, uniform: str, value, transpose=True):
        self.set_uniform(uniform, value, transpose=True)

    def set_uniform(self, name, value, transpose=True):
        if name in self._uniforms:
            typ, loc = self._uniforms[name]
            if typ in (gl.GL_FLOAT_MAT2, gl.GL_FLOAT_MAT3, gl.GL_FLOAT_MAT4):
                gl_set_uniform[typ](loc, 1, transpose, value)
            else:
                gl_set_uniform[typ](loc, 1, value)
        # else:
        #     logging.warning(f'Uniform {name} not existed.')

    def get_uniform(self, name):
        if name in self._uniforms:
            typ, loc = self._uniforms[name]
            type_info = gl_type_info[typ]
            buffer = np.zeros(type_info[0], dtype=type_info[2])
            gl_get_uniform[typ](self._id, loc, buffer.nbytes, buffer.data)
            return buffer

    def active_texture_unit(self, index: int):
        if self.is_valid():
            gl.glActiveTexture(gl.GL_TEXTURE0 + index)

    def active_next_texture_unit(self):
        if self.is_valid():
            self._current_activated_textured_unit += 1
            gl.glActiveTexture(gl.GL_TEXTURE0 + self._current_activated_textured_unit)

    @property
    def uniforms(self):
        return self._uniforms

    def _delete(self):
        gl.glDeleteProgram(self._id)

    def _activate(self):
        self.use()

    def _deactivate(self):
        self.unuse()

    @property
    def is_active(self):
        return self._is_active

    @staticmethod
    def _create_and_link(shaders: Sequence[Shader]):
        program_id = gl.glCreateProgram()
        for shader in shaders:
            gl.glAttachShader(program_id, shader.handle)
        gl.glLinkProgram(program_id)

        if not check_link_status(program_id):
            msg = f'Shader program link failure : {gl.glGetProgramInfoLog(program_id)}'
            raise RuntimeError(msg)

        # MacOs requires that the vao is bound before checking validity.
        if platform.system() != 'Darwin':
            if not check_validity(program_id):
                msg = f'Shader program validation failure : {gl.glGetProgramInfoLog(program_id)}'
                raise RuntimeError(msg)

        for shader in shaders:
            gl.glDetachShader(program_id, shader.handle)
        return program_id

    def use(self):
        gl.glUseProgram(self._id)
        self._is_active = True

    def unuse(self):
        gl.glUseProgram(0)
        self._is_active = False



from __future__ import annotations

import atexit
import logging
import os
import platform
import sys
import time

import glfw
import imgui
import numpy as np
from imgui.integrations.glfw import GlfwRenderer

from . import event
from .input import KeyCode, MouseButton, KeyModifier
from .. import misc, gl
from ..camera import Camera
from ..math import Vec2

# temporary work around to have access of OpenGL context before the app is initialised
__current_window__: Window


# TODO: detailed description of event listener parameters

class Window(event.EventDispatcher):
    """A thin wrapper around GLFW window.

    Attributes:
        _width (int): width of the window
        _height (int): height of the window
        _native_window (GLFWwindow): inner GLFWwindow

    Events:
        on_cursor_enter()
        on_cursor_leave()
        on_draw(dt)
        on_resize(width, height)
        on_mouse_motion(x, y, dx, dy)
        on_mouse_drag(x, y, dx, dy, button)
        on_mouse_press(x, y, button)
        on_mouse_release(x, y, button)
        on_mouse_scroll(x, y, x_offset, y_offset)
        on_key_press(key, mods)
        on_key_release(key, mods)
        on_update(dt)
        on_init()
        on_idle()
        on_show()
        on_hide()
        on_close()
        on_gui()
    """

    def __init__(self, title, width=600, height=600, **kwargs):
        """ Initialization of the GLFW window.

        Window creation hints are passed through `**kwargs`.

        Args:
            width (int): width of the window
            height (int): height of the window
            title (str): title of the window
            **kwargs:
                clear_color (array-like):
                    Specifies the color used to clear OpenGL color buffer.

                resizable (bool):
                    Specifies whether the window is resizable by the user.

                decorated (bool):
                    Specifies whether the windowed mode window will have window decorations such as a border,
                    a close widget, etc.

                floating (bool):
                    Specifies whether the windowed mode window will be floating above other regular windows.

                maximised (bool):
                    Specifies whether the windowed mode window will be maximized when created.

                gl_version ((int, int)):
                    OpenGL context version in tuple (major, minor).

                gl_forward_compat:
                    Specifies whether the OpenGL context should be forward-compatible, i.e. one where
                    all functionality deprecated in the requested version of OpenGL is removed. This
                    must only be used if the requested OpenGL version is 3.0 or above.

                gl_profile (str):
                    Specifies which OpenGL profile to create the context for. Possible values are
                    ['core', 'compat', 'any']

                macos_retina (bool):
                    Specifies whether to use full resolution framebuffer on Retina displays.

                macos_gfx_switch (bool):
                    Specifies whether to in Automatic Graphics Switching, i.e. to allow the system
                    to choose the integrated GPU for the OpenGL context and move it between GPUs if
                    necessary or whether to force it to always run on the discrete GPU.
        """
        super(Window, self).__init__()
        self._cursor_pos = Vec2(0., 0.)
        self._mouse_btn = MouseButton.NONE
        self._width = width
        self._height = height
        self._title = title
        self._native_window = None
        self._clear_flags = gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT | gl.GL_STENCIL_BUFFER_BIT
        self._x = 0
        self._y = 0
        self._clear_color = kwargs.get('clear_color', misc.PaletteDefault.Background)
        self._camera = None
        self._frame_rate = 0

        atexit.register(self._delete)

        if not glfw.init():
            raise RuntimeError("Failed to initialize GLFW!")

        glfw.set_error_callback(self._on_glfw_error)

        # by default the window is resizable, decorated and not floating over other windows
        glfw.window_hint(glfw.RESIZABLE, kwargs.get('resizable', True))
        glfw.window_hint(glfw.DECORATED, kwargs.get('decorated', True))
        glfw.window_hint(glfw.FLOATING, kwargs.get('floating', False))

        # by default the OpenGL version is 3.3
        (gl_major, gl_minor) = kwargs.get('gl_version', (3, 3))
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, gl_major)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, gl_minor)

        # GLFW_OPENGL_FORWARD_COMPAT is activated on macos
        if gl_major >= 3 and platform.system() == 'Darwin':  # MacOs
            glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, True)

        profile = kwargs.get('profile', 'core')
        if profile not in ['core', 'compat', 'any']:
            raise RuntimeError(f'Unknown OpenGL profile: {profile}')
        # Context profiles are only defined for OpenGL version 3.2 and above
        # by default, OpenGL context is created with core profile
        if gl_major >= 3 and profile == 'core':
            glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
        elif gl_major >= 3 and profile == 'compatibility':
            glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_COMPAT_PROFILE)
        else:
            glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_ANY_PROFILE)

        # by default use full resolution on Retina display
        glfw.window_hint(glfw.COCOA_RETINA_FRAMEBUFFER, kwargs.get('macos_retina', True))

        # by default graphics switching on macos is disabled
        glfw.window_hint(glfw.COCOA_GRAPHICS_SWITCHING, kwargs.get('macos_gfx_switch', False))

        self._native_window = glfw.create_window(width, height, title, None, None)

        if not self._native_window:
            print("Failed to create GLFW window", file=sys.stderr)
            glfw.terminate()
            sys.exit()

        glfw.make_context_current(self._native_window)

        # MacOs: Check framebuffer size and window size. On retina display, they may be different.
        w, h = glfw.get_framebuffer_size(self._native_window)
        if platform.system() == 'darwin' and (w != width or h != height):
            width, height = width // 2, height // 2
            glfw.set_window_size(self._native_window, width, height)

        self._mouse_btn = MouseButton.NONE
        self._cursor_pos = [0., 0.]

        # install callbacks
        glfw.set_framebuffer_size_callback(self._native_window, self._on_glfw_framebuffer_resize)
        glfw.set_cursor_enter_callback(self._native_window, self._on_glfw_cursor_enter)
        glfw.set_window_close_callback(self._native_window, self._on_glfw_window_close)
        glfw.set_key_callback(self._native_window, self._on_glfw_key)
        glfw.set_mouse_button_callback(self._native_window, self._on_glfw_mouse_button)
        glfw.set_cursor_pos_callback(self._native_window, self._on_glfw_mouse_motion)
        glfw.set_scroll_callback(self._native_window, self._on_glfw_scroll)
        glfw.set_char_callback(self._native_window, self._on_glfw_char)

        self._width, self._height = glfw.get_framebuffer_size(self._native_window)
        self._x, self._y = glfw.get_window_pos(self._native_window)

        self._old_width, self._old_height, self._old_pos = self._width, self._height, glfw.get_window_pos(
            self._native_window)
        self._is_fullscreen = False

        imgui.create_context()
        self._gui = GlfwRenderer(self._native_window, False) if gl_major >= 3 else None

        self._default_shader = None

        if self.current_context_version >= (3, 3):
            from ..assets.manager import default_asset_mgr
            self._default_shader = default_asset_mgr.get_or_create_pipeline('default_pipeline')
            logging.info("Default shader created.")

        self._start_time = time.time()
        self._previous_time = self._start_time
        self._current_time = self._start_time

        gl.glEnable(gl.GL_DEPTH_TEST)

        global __current_window__
        __current_window__ = self

    def shut_down(self):
        glfw.destroy_window(self._native_window)

    @staticmethod
    def _delete():
        logging.info("GLFW clean up.")
        glfw.terminate()

    def _on_glfw_framebuffer_resize(self, _window, width, height):
        self._width = width
        self._height = height
        self.dispatch('on_resize', width, height)

    def _on_glfw_cursor_enter(self, _window, entered):
        if self._gui and not imgui.get_io().want_capture_mouse:
            if entered:
                self.dispatch('on_cursor_enter')
            else:
                self.dispatch('on_cursor_leave')

    def _on_glfw_window_close(self, _window):
        self.close()

    def _on_glfw_key(self, window, key, scancode, action, mods):
        keycode = KeyCode.from_glfw_keycode(key)
        modifiers = KeyModifier.from_glfw_modifiers(mods)

        if self._gui:
            self._gui.keyboard_callback(window, key, scancode, action, mods)

        if self._gui and not imgui.get_io().want_capture_keyboard:
            if action in [glfw.PRESS, glfw.REPEAT]:
                self.dispatch('on_key_press', keycode, modifiers)
            else:
                self.dispatch('on_key_release', keycode, modifiers)

    def _on_glfw_char(self, window, char):
        if self._gui:
            self._gui.char_callback(window, char)

    def _on_glfw_mouse_button(self, window, button, action, mods):
        x, y = glfw.get_cursor_pos(window)
        button = MouseButton.from_glfw_mouse_btn_code(button)
        if action == glfw.RELEASE:
            self._mouse_btn = MouseButton.NONE
            self._cursor_pos = [x, y]
            if self._gui and not imgui.get_io().want_capture_mouse:
                self.dispatch('on_mouse_release', x, y, button)
        elif action == glfw.PRESS:
            self._mouse_btn = button
            self._cursor_pos = [x, y]
            if self._gui and not imgui.get_io().want_capture_mouse:
                self.dispatch('on_mouse_press', x, y, button)

    def _on_glfw_mouse_motion(self, _window, x, y):
        dx = x - self._cursor_pos[0]
        dy = y - self._cursor_pos[1]
        self._cursor_pos = [x, y]
        if self._gui and not imgui.get_io().want_capture_mouse:
            if self._mouse_btn != MouseButton.NONE:
                self.dispatch('on_mouse_drag', x, y, dx, dy, self._mouse_btn)
            else:
                self.dispatch('on_mouse_motion', x, y, dx, dy)

    def _on_glfw_scroll(self, window, x_offset, y_offset):
        x, y = glfw.get_cursor_pos(window)
        if self._gui and not imgui.get_io().want_capture_mouse:
            self.dispatch('on_mouse_scroll', x, y, x_offset, y_offset)

    def _on_glfw_error(self, error, desc):
        print(f'GLFW Error: {desc}', file=sys.stderr)

    def toggle_fullscreen(self):
        monitor = glfw.get_primary_monitor()
        mode = glfw.get_video_mode(monitor)
        if not self._is_fullscreen:
            self._old_width, self._old_height = self.width, self.height
            self._old_pos = glfw.get_window_pos(self._native_window)
            glfw.set_window_size(self._native_window, mode.size.width, mode.size.height)
            glfw.set_window_pos(self._native_window, 0, 0)
        else:
            glfw.set_window_size(self._native_window, self._old_width, self._old_height)
            glfw.set_window_pos(self._native_window, self._old_pos[0], self._old_pos[1])

        self._is_fullscreen = not self._is_fullscreen

    @property
    def title(self):
        return self._title

    @property
    def width(self):
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)
        return self._width

    @width.setter
    def width(self, value):
        glfw.set_window_size(self._native_window, value, self._height)
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)

    @property
    def height(self):
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)
        return self._height

    @height.setter
    def height(self, value):
        glfw.set_window_size(self._native_window, self._width, self._height)
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)

    @property
    def aspect_ratio(self):
        return self._width / self._height

    @property
    def size(self):
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)
        return self._width, self._height

    @size.setter
    def size(self, sz):
        glfw.set_window_size(self._native_window, sz[0], sz[1])
        self._width, self._height = glfw.get_framebuffer_size(self._native_window)

    @property
    def position(self):
        self._x, self._y = glfw.get_window_pos(self._native_window)
        return Vec2(self._x, self._y)

    @position.setter
    def position(self, pos):
        glfw.set_window_pos(self._native_window, pos[0], pos[1])
        self._x, self._y = glfw.get_window_pos(self._native_window)

    @property
    def current_context_version(self):
        return glfw.get_window_attrib(self._native_window, glfw.CONTEXT_VERSION_MAJOR), \
               glfw.get_window_attrib(self._native_window, glfw.CONTEXT_VERSION_MINOR)

    @property
    def default_shader(self):
        return self._default_shader

    @default_shader.setter
    def default_shader(self, shader):
        self._default_shader = shader

    @property
    def delta_time(self):
        self._current_time = time.time()
        elapsed = self._current_time - self._previous_time
        self._previous_time = self._current_time
        return elapsed

    @property
    def elapsed_time(self):
        return time.time() - self._start_time

    def main_loop(self):
        """
        Start the main loop of the window.
        """
        self._previous_time = time.time()

        while not glfw.window_should_close(self._native_window):
            # update time
            dt = self.delta_time
            if dt != 0.0:
                self._frame_rate = int(1.0 / dt)

            # process inputs
            glfw.poll_events()

            if self._gui:
                self._gui.process_inputs()
                imgui.new_frame()
                imgui.set_next_window_position(self._width - 150, 30)
                imgui.set_next_window_size(120, 20)
                imgui.begin('Frame Rate', False,
                            imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE)
                imgui.text(f'framerate: {self._frame_rate}')
                imgui.end()
                if 'on_gui' in self._event_listeners and len(self._event_listeners['on_gui']) > 0:
                    imgui.begin('Controls')
                    self.dispatch('on_gui', reversed_exec_order=True)
                    imgui.end()
                imgui.end_frame()
                imgui.render()

            # update
            self.dispatch('on_update', dt)

            # render
            gl.glClear(self._clear_flags)

            self.dispatch('on_draw', dt)

            self.dispatch('on_idle', dt)

            if self._gui:
                draw_data = imgui.get_draw_data()
                if draw_data:
                    self._gui.render(draw_data)

            glfw.swap_buffers(self._native_window)

        self.dispatch('on_close')

    def clear(self, color):
        color = color if color else self._clear_color
        gl.glClearColor(*color)
        gl.glClear(self._clear_flags)

    def swap_buffers(self):
        glfw.swap_buffers(self._native_window)

    def activate(self):
        glfw.make_context_current(self._native_window)

    def destroy(self):
        glfw.destroy_window(self._native_window)

    def close(self):
        glfw.set_window_should_close(self._native_window, True)

    def show(self):
        glfw.show_window(self._native_window)
        self.dispatch('on_show')

    def hide(self):
        glfw.hide_window(self._native_window)
        self.dispatch('on_hide')

    def on_init(self):
        gl.glClearColor(*self._clear_color)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

    def on_close(self):
        if self._gui:
            self._gui.shutdown()
        glfw.terminate()

    def on_resize(self, width, height):
        """Default resize handler."""
        self.activate()
        gl.glViewport(0, 0, width, height)
        gl.glClear(self._clear_flags)
        self.dispatch('on_draw', self.delta_time)
        self.swap_buffers()

    def on_key_press(self, key, mods):
        if key == KeyCode.Escape:
            self.close()
        elif key == KeyCode.F10:
            import png
            import datetime
            h, w = self.size
            framebuffer = np.zeros((h, w * 3), dtype=np.uint8)
            gl.glReadPixels(0, 0, w, h, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, framebuffer)
            filename = f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.png'
            filepath = os.path.join(os.getcwd(), filename)
            png.from_array(framebuffer[::-1], 'RGB').save(filepath)
            print(f'Screenshot saved to {filepath}')
        elif key == KeyCode.F11:
            self.toggle_fullscreen()
            self.dispatch('on_resize', self.width, self.height)

        return True

    def mouse_position(self):
        return glfw.get_cursor_pos(self._native_window)

    def create_camera(self, pos, look_at, up, fov_v=45.0, near=0.1, far=1000., degrees=True, zoom_enabled=False,
                      safe_rotations=True, zoom_speed=0.25):
        self._camera = Camera(pos, look_at, up, self._width / self._height, fov_v, near, far, degrees, zoom_enabled,
                              safe_rotations, zoom_speed)
        self.attach_listeners(self._camera)

    @property
    def camera(self):
        return self._camera


Window.register_event_type('on_cursor_enter')
Window.register_event_type('on_cursor_leave')
Window.register_event_type('on_draw')
Window.register_event_type('on_resize')
Window.register_event_type('on_mouse_motion')
Window.register_event_type('on_mouse_drag')
Window.register_event_type('on_mouse_press')
Window.register_event_type('on_mouse_release')
Window.register_event_type('on_mouse_scroll')
Window.register_event_type('on_key_press')
Window.register_event_type('on_key_release')
Window.register_event_type('on_update')
Window.register_event_type('on_init')
Window.register_event_type('on_idle')
Window.register_event_type('on_show')
Window.register_event_type('on_hide')
Window.register_event_type('on_close')
Window.register_event_type('on_gui')


class WindowEventLogger:
    """Logs window events when triggered.
    """

    def __init__(self, log_file=None):
        """Creates a `WindowEventLogger` object.

        Args:
            log_file (file like object):
                Specifies the output of log messages. If not specified, stdout will be used.
        """
        self._output = log_file if log_file is not None else sys.stdout

    def on_key_press(self, key, mods):
        print(f'on_key_press({key}, mods={mods})', file=self._output)

    def on_key_release(self, key, mods):
        print(f'on_key_release({key}, mods={mods})', file=self._output)

    def on_mouse_motion(self, x, y, dx, dy):
        print(f'on_mouse_motion(x={x}, y={y}, dx={dx}, dy={dy})', file=self._output)

    def on_mouse_drag(self, x, y, dx, dy, btns, mods):
        print(f'on_mouse_drag(x={x}, y={y}, dx={dx}, dy={dy}, btns={btns}, mods={mods}', file=self._output)

    def on_mouse_press(self, x, y, btn, mods):
        print(f'on_mouse_press(x={x}, y={y}, btn={btn}, mods={mods}', file=self._output)

    def on_mouse_release(self, x, y, btn, mods):
        print(f'on_mouse_release(x={x}, y={y}, btn={btn}, mods={mods}', file=self._output)

    def on_mouse_scroll(self, x, y, dx, dy):
        print(f'on_mouse_scroll(x={x}, y={y}, dx={dx}, dy={dy})', file=self._output)

    def on_close(self):
        print(f'on_close()', file=self._output)

    def on_resize(self, w, h):
        print(f'on_resize(w={w}, h={h})', file=self._output)

    def on_draw(self):
        print(f'on_draw()', file=self._output)



# from __future__ import annotations

import math

import OpenGL.GL as gl

from .. import app
from ..app.input import MouseButton
from ..math import Vec3, Vec4, Mat4


# TODO: decouple camera and control


class Camera:
    def __init__(self, pos, look_at, up, aspect_ratio, fov_v=45.0, near=0.1, far=1000., degrees=True,
                 zoom_enabled=False, safe_rotations=True, zoom_speed=0.25):
        self._pos = Vec3(pos)
        self._look_at = Vec3(look_at)
        self._up = Vec3(up)
        self._aspect_ratio = aspect_ratio
        self._near = near
        self._far = far
        self._fov = fov_v if not degrees else math.radians(fov_v)
        self._proj_mat = Mat4.perspective_gl(self._fov, self._aspect_ratio, self._near, self._far)
        self._view_mat = Mat4.look_at_gl(self._pos, self._look_at, self._up)
        self._d_angle_x = 0.
        self._d_angle_y = 0.
        self._zoom_enabled = zoom_enabled
        self._safe_rotations = safe_rotations
        self._zoom_speed = zoom_speed

    @property
    def position(self):
        return self._pos

    @property
    def field_of_view(self):
        return self._fov

    @field_of_view.setter
    def field_of_view(self, value):
        raise NotImplementedError

    @property
    def aspect_ratio(self):
        return self._aspect_ratio

    @aspect_ratio.setter
    def aspect_ratio(self, value):
        self._aspect_ratio = value
        self._update_matrices()

    @property
    def near(self):
        return self._near

    @property
    def far(self):
        return self._far

    @property
    def forward(self):
        return Vec3(-self._view_mat.row(2))

    @property
    def right(self):
        return Vec3(self._view_mat.row(0))

    @property
    def up(self) -> Vec3:
        return self._up

    @property
    def projection_matrix(self) -> Mat4:
        return self._proj_mat

    @property
    def view_matrix(self) -> Mat4:
        return self._view_mat

    @property
    def look_at(self) -> Vec3:
        return self._look_at

    @property
    def zoom_enabled(self):
        return self._zoom_enabled

    @zoom_enabled.setter
    def zoom_enabled(self, value):
        self._zoom_enabled = value

    @property
    def safe_rotations(self):
        return self._safe_rotations

    @safe_rotations.setter
    def safe_rotations(self, value):
        self._safe_rotations = value

    def __enter__(self):
        self._update_matrices()
        return self

    def __exit__(self, *args):
        pass

    def update_view(self, eye, look_at, up):
        self._pos = Vec3(eye)
        self._look_at = Vec3(look_at)
        self._up = Vec3(up)
        self._update_matrices()

    def _update_matrices(self):
        self._proj_mat = Mat4.perspective_gl(self._fov, self._aspect_ratio, self._near, self._far)
        self._view_mat = Mat4.look_at_gl(self._pos, self._look_at, self._up)

    def on_update(self, dt):
        shader = app.current_window().default_shader
        if shader is not None:
            with shader:
                # logging.info(f'on_update of {self}')
                shader.view_mat = self._view_mat
                shader.proj_mat = self._proj_mat
        else:
            gl.glMatrixMode(gl.GL_PROJECTION)
            gl.glLoadTransposeMatrixf(self._proj_mat)

            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadTransposeMatrixf(self._view_mat)

    def on_resize(self, width, height):
        self._d_angle_x = 2.0 * math.pi / width if width != 0 else 0.0
        self._d_angle_y = math.pi / height
        self._aspect_ratio = width / height
        self._update_matrices()

    def on_mouse_drag(self, x, y, dx, dy, btn):
        if btn == MouseButton.Left:
            pos = Vec4(self._pos, 1.0)
            pivot = Vec4(self._look_at, 1.0)
            angle_x = dx * self._d_angle_x
            angle_y = dy * self._d_angle_y
            cos_angle = self.forward.dot(self.up)
            # When the camera direction is the same as the up vector
            if cos_angle > 0.99:
                self._up = Vec3(Mat4.from_axis_angle(self.right, 45, True) * Vec4(self._up, 1.0))
            elif cos_angle < -0.99:
                self._up = Vec3(Mat4.from_axis_angle(-self.right, 45, True) * Vec4(self._up, 1.0))

            # Rotate the camera around the pivot point.
            rot_x = Mat4.from_axis_angle(self.up, angle_x)
            pos = rot_x * (pos - pivot) + pivot
            rot_y = Mat4.from_axis_angle(self.right, angle_y)
            pos = rot_y * (pos - pivot) + pivot

            new_cos_angle = Vec3(pos).normalise().dot(self.up)
            if not (self.safe_rotations and abs(new_cos_angle) > 0.9):
                self.update_view(pos, self.look_at, self.up)

    def on_mouse_scroll(self, x, y, x_offset, y_offset, min_dist=1e-5):
        if self._zoom_enabled:
            look_dir = (self.look_at - self.position).normalise()
            pos = self.position + look_dir * y_offset * self._zoom_speed
            if pos.norm > min_dist:
                self.update_view(pos, self.look_at, self.up)



import numbers
from typing import Sequence

import numpy as np

from .shape import Shape
from .ray import Ray
from ..graphics.util import DrawingMode
from ..math import Vec3
from ..misc import Color, PaletteDefault


class Triangle(Shape):
    def __init__(self, p0, p1, p2, colors: Sequence[Color] = (PaletteDefault.YellowB.as_color(),)):
        """Creates a triangle from three positions and colors.

        Args:
            p0 (array_like, shape (3,) or list):
                First point of the triangle.

            p1 (array_like, shape (3,) or list):
                Second point of the triangle.

            p2 (array_like, shape (3,) or list):
                Third point of the triangle.

            colors (Sequence[Color]):
                Specifies the color of each point of triangle.
        """
        super(Triangle, self).__init__(3, colors)
        self._points = np.array([p0, p1, p2]).reshape((3, 3))

    def __str__(self):
        _str = 'Triangle [{}, {}, {}]'
        return _str.format(*self._points)

    def __getitem__(self, item: int):
        if not isinstance(item, numbers.Integral) or item < 0 or item > 2:
            raise IndexError(f'Invalid index {item}')
        return self._points[item].reshape((3, 1)).view(Vec3)

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i < 3:
            value = self._points[self._i]
            self._i += 1
            return value
        else:
            raise StopIteration

    @property
    def vertex_count(self) -> int:
        return 3

    @property
    def drawing_mode(self) -> DrawingMode:
        return DrawingMode.Triangles

    @property
    def p0(self) -> Vec3:
        return self._points[0].reshape(3, 1).view(Vec3)

    @property
    def p1(self) -> Vec3:
        return self._points[1].reshape(3, 1).view(Vec3)

    @property
    def p2(self) -> Vec3:
        return self._points[2].reshape(3, 1).view(Vec3)

    @property
    def vertices(self) -> np.ndarray:
        return np.array(self._points).ravel()

    def intersect_with_ray(self, ray: Ray) -> bool:
        a = np.array([[self[0].x - self[1].x, self[0].x - self[2].x, ray.direction.x],
                      [self[0].y - self[1].y, self[0].y - self[2].y, ray.direction.y],
                      [self[0].z - self[1].z, self[0].z - self[2].z, ray.direction.z]])

        b = np.array([self[0].x - ray.origin.x, self[0].y - ray.origin.y, self[0].z - ray.origin.z])

        x = np.linalg.solve(a, b)

        if x[0] < 0 or x[1] < 0 or x[0] + x[1] > 1 or x[2] < 0:
            return False

        return True

    @property
    def indices(self):
        return np.array([0, 1, 2], dtype=int)

    @property
    def index_count(self):
        return 3



import abc
import uuid


class Entity(metaclass=abc.ABCMeta):
    def __init__(self, name=None, cast_shadow=True):
        self._uuid = uuid.uuid1()
        self._name = str(name) if name else 'Unnamed{}'.format(self.__class__.__name__)
        self._is_drawable = True
        self._is_visible = True
        self._cast_shadow = cast_shadow

    @property
    def visible(self):
        return self._is_visible

    @visible.setter
    def visible(self, value):
        self._is_visible = value

    @property
    def drawable(self):
        return self._is_drawable

    @drawable.setter
    def drawable(self, value):
        self._is_drawable = value

    @property
    def cast_shadow(self):
        return self._cast_shadow

    @cast_shadow.setter
    def cast_shadow(self, value):
        self._cast_shadow = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def uuid(self):
        return self._uuid

    @property
    @abc.abstractmethod
    def meshes(self):
        return NotImplementedError




import enum
import math

import numpy as np

from .shape import Shape
from ..graphics.util import DrawingMode
from ..math import Vec3, Vec2
from ..misc import PaletteDefault


class AxisAlignment(enum.Enum):
    XY = 0
    XZ = 1
    YX = 2
    YZ = 3
    ZX = 4
    ZY = 5


class Grid(Shape):
    def __init__(self,
                 width=20.0,
                 height=20.0,
                 spacing_width=1.0,
                 spacing_height=1.0,
                 origin=Vec3(0.0),
                 axis_alignment=AxisAlignment.XZ,
                 axis_marker=False):
        points, colors = self._generate_lines(origin, width, height, spacing_width, spacing_height, axis_alignment,
                                              axis_marker)
        super().__init__(len(points), colors)
        self._origin = origin
        self._width = width
        self._height = height
        self._spacing = Vec2(spacing_width, spacing_height)
        self._points = points
        self._colors = colors

    @staticmethod
    def _generate_lines(o, w, h, dw, dh, alignment, axis_marker):
        lines = []
        colors = []
        count_w = math.ceil(w / dw)
        count_h = math.ceil(h / dh)
        color_axis_vertical = PaletteDefault.Background
        for w_i in range(0, count_w + 1):
            a = -w / 2 + w_i * dw
            b = h / 2
            if alignment is AxisAlignment.XZ:
                lines.extend([Vec3(a, 0.0, -b) + o,
                              Vec3(a, 0.0, b) + o])
                color_axis_vertical = PaletteDefault.BlueA.as_color()
            elif alignment is AxisAlignment.XY:
                lines.extend([Vec3(a, -b, 0.0) + o,
                              Vec3(a, b, 0.0) + o])
                color_axis_vertical = PaletteDefault.GreenA.as_color()
            elif alignment is AxisAlignment.YX:
                lines.extend([Vec3(-b, a, 0.0) + o,
                              Vec3(b, a, 0.0) + o])
                color_axis_vertical = PaletteDefault.RedA.as_color()
            elif alignment is AxisAlignment.YZ:
                lines.extend([Vec3(0.0, a, -b) + o,
                              Vec3(0.0, a, b) + o])
                color_axis_vertical = PaletteDefault.BlueA.as_color()
            elif alignment is AxisAlignment.ZX:
                lines.extend([Vec3(-b, 0.0, a) + o,
                              Vec3(b, 0.0, a) + o])
                color_axis_vertical = PaletteDefault.RedA.as_color()
            elif alignment is AxisAlignment.ZY:
                lines.extend([Vec3(0.0, -b, a) + o,
                              Vec3(0.0, b, a) + o])
                color_axis_vertical = PaletteDefault.GreenA.as_color()
            else:
                pass
            if axis_marker and w_i == w // 2:
                colors.append(color_axis_vertical)
                colors.append(color_axis_vertical)
            else:
                colors.append(PaletteDefault.BrownB.as_color())
                colors.append(PaletteDefault.BrownB.as_color())
        color_axis_horizontal = PaletteDefault.Background
        for h_i in range(0, count_h + 1):
            a = w / 2
            b = -h / 2 + h_i * dh
            if alignment == AxisAlignment.XZ:
                lines.extend([Vec3(-a, 0.0, b) + o,
                              Vec3(a, 0.0, b) + o])
                color_axis_horizontal = PaletteDefault.RedA.as_color()
            elif alignment is AxisAlignment.XY:
                lines.extend([Vec3(-a, b, 0.0) + o,
                              Vec3(a, b, 0.0) + o])
                color_axis_horizontal = PaletteDefault.RedA.as_color()
            elif alignment is AxisAlignment.YX:
                lines.extend([Vec3(b, -a, 0.0) + o,
                              Vec3(b, a, 0.0) + o])
                color_axis_horizontal = PaletteDefault.GreenA.as_color()

            elif alignment is AxisAlignment.YZ:
                lines.extend([Vec3(0.0, -a, b) + o,
                              Vec3(0.0, a, b) + o])
                color_axis_horizontal = PaletteDefault.GreenA.as_color()

            elif alignment is AxisAlignment.ZX:
                lines.extend([Vec3(b, 0.0, -a) + o,
                              Vec3(b, 0.0, a) + o])
                color_axis_horizontal = PaletteDefault.BlueA.as_color()

            elif alignment is AxisAlignment.ZY:
                lines.extend([Vec3(0.0, b, -a) + o,
                              Vec3(0.0, b, a) + o])
                color_axis_horizontal = PaletteDefault.BlueA.as_color()
            else:
                pass
            if axis_marker and h_i == w // 2:
                colors.append(color_axis_horizontal)
                colors.append(color_axis_horizontal)
            else:
                colors.append(PaletteDefault.BrownB.as_color())
                colors.append(PaletteDefault.BrownB.as_color())

        return np.asarray(lines, dtype=np.float32).reshape((-1, 3)), colors

    @property
    def vertices(self) -> np.ndarray:
        return self._points.ravel()

    @property
    def vertex_count(self) -> int:
        return len(self._points)

    @property
    def indices(self):
        return np.array(list(range(0, len(self._points))), dtype=np.uint32)

    @property
    def index_count(self):
        return len(self._points)

    @property
    def drawing_mode(self) -> DrawingMode:
        return DrawingMode.Lines



import datetime
import enum
import logging
import math
import os
from collections.abc import Sequence

import numpy as np
import png

from .texture import TextureWrapMode
from .. import gl
from .util import GpuObject, BindSemanticObject


"""
An attachment can be a texture or a renderbuffer. Main difference between these two is that as for a texture,
you can access it from shader, but for renderbuffer you can only access its content using glReadPixels.

Renderbuffer Objects are OpenGL objects that contain images. They are created and used specifically with 
Framebuffer Objects. They are optimized for use as render targets, while Textures may not be, and are the
logical choice when you do not need to sample (i.e. in a post-pass shader) from the produced image. If 
you need to resample (such as when reading depth back in a second shader pass), use Textures instead.
Renderbuffer objects natively accommodate multisampling (MSAA).
"""


class AttachmentKind(enum.Enum):
    Color = 0
    Depth = 1
    Stencil = 2
    DepthStencil = 3


class Attachment(GpuObject, BindSemanticObject):
    """An attachment is an image (texture)."""
    def __init__(self, width, height, kind: AttachmentKind, internal_fmt, fmt, dtype, shader_accessible=True, **kwargs):
        """

        Args:
            width:
            height:
            kind:
            internal_fmt (OpenGL base internal formats, sized internal formats or compressed internal formats):
                Specifies the number of color components in the texture. Only needed if shader_accessible is true.
            fmt:
                Specifies the format of the pixel data. Accepted values: GL_RED, GL_RG, GL_RGB, GL_BGR, GL_RGBA,
                GL_BGRA, GL_RED_INTEGER, GL_RG_INTEGER, GL_RGB_INTEGER, GL_BGR_INTEGER, GL_RGBA_INTEGER,
                GL_BGRA_INTEGER, GL_STENCIL_INDEX, GL_DEPTH_COMPONENT, GL_DEPTH_STENCIL.
            dtype:
                Specifies the data type of the pixel data. Accepted values: GL_UNSIGNED_BYTE, GL_BYTE,
                GL_UNSIGNED_SHORT, GL_SHORT, GL_UNSIGNED_INT, GL_INT, GL_HALF_FLOAT, GL_FLOAT, GL_UNSIGNED_BYTE_3_3_2,
                GL_UNSIGNED_BYTE_2_3_3_REV, GL_UNSIGNED_SHORT_5_6_5, GL_UNSIGNED_SHORT_5_6_5_REV,
                GL_UNSIGNED_SHORT_4_4_4_4, GL_UNSIGNED_SHORT_4_4_4_4_REV, GL_UNSIGNED_SHORT_5_5_5_1,
                GL_UNSIGNED_SHORT_1_5_5_5_REV, GL_UNSIGNED_INT_8_8_8_8, GL_UNSIGNED_INT_8_8_8_8_REV,
                 GL_UNSIGNED_INT_10_10_10_2, and GL_UNSIGNED_INT_2_10_10_10_REV.
            shader_accessible:
        """
        target = gl.GL_TEXTURE_2D if shader_accessible else gl.GL_RENDERBUFFER
        super().__init__(target, -1)
        self._kind = kind
        self._target = target
        self._is_shader_accessible = shader_accessible
        self._width = width
        self._height = height
        self._internal_format = internal_fmt
        self._format = fmt
        self._dtype = dtype
        self._wrap_mode = kwargs.get('wrap_mode', TextureWrapMode.ClampBorder)
        self._border_color = kwargs.get('border_color', (1.0, 1.0, 1.0))
        self._is_dirty = True

        self._create()

    def _create(self):
        logging.info(f'Framebuffer attachment creation <{"texture" if self._is_shader_accessible else "renderbuffer"}>')
        if self._is_shader_accessible:
            self._id = gl.glGenTextures(1)
        else:
            self._id = gl.glGenRenderbuffers(1)

    def _delete(self):
        if self.is_valid():
            if self._is_shader_accessible:
                gl.glDeleteTextures(1, [self._id])
            else:
                gl.glDeleteRenderbuffers(1, [self._id])

    def _activate(self):
        self.bind()

    def _deactivate(self):
        self.unbind()

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def kind(self) -> AttachmentKind:
        return self._kind

    @property
    def is_shader_accessible(self):
        return self._is_shader_accessible

    @property
    def target(self):
        return self._target

    def resize(self, width, height):
        if width != self._width or height != self._height:
            self._is_dirty = True
            self._width = width
            self._height = height

    def bind(self):
        """Activate the attachment."""
        if self.is_valid():
            if self._is_shader_accessible:
                gl.glBindTexture(self._target, self._id)
            else:
                gl.glBindRenderbuffer(self._target, self._id)

            if self._is_dirty:
                self._resize()
                self._is_dirty = False

    def unbind(self):
        """Deactivate the attachment."""
        if self._is_shader_accessible:
            gl.glBindTexture(self._target, 0)
        else:
            gl.glBindRenderbuffer(self._target, 0)

    def _resize(self):
        if self._is_shader_accessible:
            gl.glTexImage2D(self._target, 0, self._internal_format, self._width, self._height, 0, self._format,
                            self._dtype, None)
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_S, self._wrap_mode.value)
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_T, self._wrap_mode.value)
            gl.glTexParameterfv(self._target, gl.GL_TEXTURE_BORDER_COLOR, self._border_color)
        else:
            gl.glRenderbufferStorage(self._target, self._format, self._width, self._height)


class ColorAttachment(Attachment):
    def __init__(self, width, height, internal_fmt=gl.GL_RGBA, fmt=gl.GL_RGBA,
                 dtype=gl.GL_UNSIGNED_BYTE, shader_accessible=True):
        super().__init__(width, height, AttachmentKind.Color, internal_fmt, fmt, dtype, shader_accessible)


class DepthAttachment(Attachment):
    def __init__(self, width, height, internal_fmt=gl.GL_DEPTH_COMPONENT, fmt=gl.GL_DEPTH_COMPONENT,
                 dtype=gl.GL_FLOAT, shader_accessible=False):
        super().__init__(width, height, AttachmentKind.Depth, internal_fmt, fmt, dtype, shader_accessible)


class StencilAttachment(Attachment):
    def __init__(self, width, height, internal_fmt=gl.GL_STENCIL_INDEX, fmt=gl.GL_STENCIL_INDEX,
                 dtype=gl.GL_UNSIGNED_INT, shader_accessible=False):
        super().__init__(width, height, AttachmentKind.Stencil, internal_fmt, fmt, dtype, shader_accessible)


class DepthStencilAttachment(Attachment):
    def __init__(self, width, height, internal_fmt=gl.GL_DEPTH24_STENCIL8, fmt=gl.GL_DEPTH24_STENCIL8,
                 dtype=gl.GL_UNSIGNED_INT_24_8, shader_accessible=False):
        super().__init__(width, height, AttachmentKind.Stencil, internal_fmt, fmt, dtype, shader_accessible)


_color_attachment = [
    gl.GL_COLOR_ATTACHMENT0,
    gl.GL_COLOR_ATTACHMENT1,
    gl.GL_COLOR_ATTACHMENT2,
    gl.GL_COLOR_ATTACHMENT3,
    gl.GL_COLOR_ATTACHMENT4,
    gl.GL_COLOR_ATTACHMENT5,
    gl.GL_COLOR_ATTACHMENT6,
    gl.GL_COLOR_ATTACHMENT7,
    gl.GL_COLOR_ATTACHMENT8,
    gl.GL_COLOR_ATTACHMENT9,
    gl.GL_COLOR_ATTACHMENT10,
    gl.GL_COLOR_ATTACHMENT11,
    gl.GL_COLOR_ATTACHMENT12,
    gl.GL_COLOR_ATTACHMENT13,
    gl.GL_COLOR_ATTACHMENT14,
    gl.GL_COLOR_ATTACHMENT15,
    gl.GL_COLOR_ATTACHMENT16,
    gl.GL_COLOR_ATTACHMENT17,
    gl.GL_COLOR_ATTACHMENT18,
    gl.GL_COLOR_ATTACHMENT19,
    gl.GL_COLOR_ATTACHMENT20,
    gl.GL_COLOR_ATTACHMENT21,
    gl.GL_COLOR_ATTACHMENT22,
    gl.GL_COLOR_ATTACHMENT23,
    gl.GL_COLOR_ATTACHMENT24,
    gl.GL_COLOR_ATTACHMENT25,
    gl.GL_COLOR_ATTACHMENT26,
    gl.GL_COLOR_ATTACHMENT27,
    gl.GL_COLOR_ATTACHMENT28,
    gl.GL_COLOR_ATTACHMENT29,
    gl.GL_COLOR_ATTACHMENT30,
    gl.GL_COLOR_ATTACHMENT31,
]


class Framebuffer(GpuObject, BindSemanticObject):
    """
    A Framebuffer object is a collection of attachments (images).
    Attachment points:
      - color attachment
      - depth attachment
      - stencil attachment
      - depth stencil attachment
    """
    def __init__(self, width, height,
                 color: Sequence[ColorAttachment] = (),
                 depth: Sequence[DepthAttachment] = (),
                 stencil: Sequence[StencilAttachment] = (), **kwargs):
        """

        Args:
            width:
            height:
            color:
            depth:
            stencil:
            **kwargs:
                enable_color_attachment (bool) : Defaults to True.
                enable_depth_attachment (bool) : Defaults to True.
                enable_stencil_attachment (bool) : Defaults to False.
                color_shader_accessible (bool) : Defaults to True.
                depth_shader_accessible (bool) : Defaults to True.
                stencil_shader_accessible (bool) : Defaults to True.
        """
        super().__init__(gl.GL_FRAMEBUFFER, -1)
        self._target = gl.GL_FRAMEBUFFER
        self._width = width
        self._height = height

        self._color_enabled = kwargs.get('enabled_color_attachment', True)
        shader_accessible = kwargs.get('color_shader_accessible', True)
        if self._color_enabled:
            self._color = [ColorAttachment(width, height, shader_accessible=shader_accessible)] if len(color) == 0 else list(color)
        else:
            self._color = []

        self._depth_enabled = kwargs.get('enable_depth_attachment', True)
        shader_accessible = kwargs.get('depth_shader_accessible', False)
        if self._depth_enabled:
            self._depth = [DepthAttachment(width, height, shader_accessible=shader_accessible)] if len(depth) == 0 else list(depth)
        else:
            self._depth = []

        self._stencil_enabled = kwargs.get('enable_stencil_attachment', False)
        shader_accessible = kwargs.get('stencil_shader_accessible', False)
        if self._stencil_enabled:
            self._stencil = [StencilAttachment(width, height, shader_accessible=shader_accessible)] if len(stencil) == 0 else list(stencil)
        else:
            self._stencil = []

        self._is_activated = False

        # Creation of framebuffer object
        self._id = gl.glGenFramebuffers(1)

        # Bind the framebuffer to activate it
        gl.glBindFramebuffer(self._target, self._id)

        # Attach buffers
        if len(self._color) > 0:
            for i, color in enumerate(self._color):
                with color:  # bind to use
                    if color.is_shader_accessible:
                        gl.glFramebufferTexture2D(self._target, _color_attachment[i], color.target, color.handle, 0)
                    else:
                        gl.glFramebufferRenderbuffer(self._target, _color_attachment[i], color.target, color.handle)
        if len(self._depth) > 0:
            for i, depth in enumerate(self._depth):
                with depth:  # bind to use
                    if depth.is_shader_accessible:
                        gl.glFramebufferTexture2D(self._target, gl.GL_DEPTH_ATTACHMENT, depth.target, depth.handle, 0)
                    else:
                        gl.glFramebufferRenderbuffer(self._target, gl.GL_DEPTH_ATTACHMENT, depth.target, depth.handle)
        if len(self._stencil) > 0:
            for i, stencil in enumerate(self._stencil):
                with stencil:  # bind to use
                    if stencil.is_shader_accessible:
                        gl.glFramebufferTexture2D(self._target, gl.GL_STENCIL_ATTACHMENT, stencil.target, stencil.handle, 0)
                    else:
                        gl.glFramebufferRenderbuffer(self._target, gl.GL_STENCIL_ATTACHMENT, stencil.target, stencil.handle)

        if gl.glCheckFramebufferStatus(self._target) != gl.GL_FRAMEBUFFER_COMPLETE:
            raise ValueError('Framebuffer is not complete!')

        gl.glBindFramebuffer(self._target, 0)

    def _activate(self):
        self.bind()

    def _deactivate(self):
        self.unbind()

    def bind(self):
        if self.is_valid():
            gl.glBindFramebuffer(self._target, self._id)
            self._is_activated = True

    def unbind(self):
        if self.is_valid():
            gl.glBindFramebuffer(self._target, 0)
            self._is_activated = False

    def _delete(self):
        logging.info("GPU: delete framebuffer.")
        gl.glDeleteFramebuffers(1, [self._id])

    @property
    def color_attachments(self):
        return self._color

    @color_attachments.setter
    def color_attachments(self, buffers):
        pass

    @property
    def depth_attachments(self):
        return self._depth

    @depth_attachments.setter
    def depth_attachments(self, buffers):
        pass

    @property
    def stencil_attachments(self):
        return self._stencil

    @stencil_attachments.setter
    def stencil_attachments(self, buffers):
        pass

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def clear(self, color=(0.635, 0.824, 1.0, 1.0), color_bit=True, depth_bit=True, stencil_bit=True):
        gl.glClearColor(*color)
        if self._is_activated:
            flags = 0
            if color_bit:
                flags |= gl.GL_COLOR_BUFFER_BIT
            if depth_bit:
                flags |= gl.GL_DEPTH_BUFFER_BIT
            if stencil_bit:
                flags |= gl.GL_STENCIL_BUFFER_BIT

            gl.glClear(flags)

    def enable_depth_test(self):
        if self._is_activated:
            gl.glEnable(gl.GL_DEPTH_TEST)

    def enable_stencil_test(self):
        if self._is_activated:
            gl.glEnable(gl.GL_STENCIL_TEST)

    def disable_color_output(self):
        """
        Tells OpenGL that we are not going to render any color data.
        """
        if self._is_activated:
            gl.glDrawBuffer(gl.GL_NONE)
            gl.glReadBuffer(gl.GL_NONE)

    def enable_color_output(self, index):
        if self._is_activated:
            if index < len(self._color):
                gl.glDrawBuffer(_color_attachment[index])
                gl.glReadBuffer(_color_attachment[index])

    def resize(self, width, height):
        if self._color:
            for c in self._color:
                c.resize(width, height)
        if self._depth:
            for d in self._depth:
                d.resize(width, height)
        if self._stencil:
            for s in self._stencil:
                s.resize(width, height)

    def save_color_attachment(self, filename=None, save_as_image=True):
        if self._color_enabled:
            with self:
                data = np.zeros((self._height, self._width * 4), dtype=np.uint8)
                gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0)
                gl.glReadPixels(0, 0, self._width, self._height, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, data)
                if save_as_image:
                    filename = f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}_color_attach.png' if filename is None else filename
                    filepath = os.path.join(os.getcwd(), filename)
                    png.from_array(data[::-1], 'RGBA').save(filepath)
                    print(f'Color attachment saved to {filepath}')
                return data

    def save_depth_attachment(self, near, far, is_perspective=True, filename=None, save_as_image=True):
        if self._depth_enabled:
            with self:
                data = np.zeros((self._height, self._width), dtype=np.float32)
                gl.glReadPixels(0, 0, self._width, self._height, gl.GL_DEPTH_COMPONENT, gl.GL_FLOAT, data)
                data = np.vectorize(Framebuffer._depth_value_normalisation)(data, near, far, is_perspective).astype(np.uint8)
                if save_as_image:
                    filename = f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}_depth_attach.png' if filename is None else filename
                    filepath = os.path.join(os.getcwd(), filename)
                    png.from_array(data[::-1], 'L').save(filepath)
                    print(f'Depth attachment saved to {filepath}')
                return data

    @staticmethod
    def _depth_value_normalisation(d, near, far, is_perspective):
        if is_perspective:
            # from range [0, 1] to NDC in range [-1, 1]
            ndc = d * 2.0 - 1.0
            # inverse the transformation to retrieve linear depth value
            value = (2.0 * near * far) / (far + near - ndc * (far - near))
            # now the value is between near and far
            value /= far - near
            # remap to [0, 255]
            value = math.floor(value * 255.0)
        else:
            value = math.floor(d * 255.0)
        return value

    def save_stencil_attachment(self):
        pass



from .light import Light
from ...math import Vec3, Mat4
from ...misc import PaletteSvg, Color


class PointLight(Light):
    def __init__(self, position: Vec3 = Vec3(5.0, 5.0, 5.0), color: Color = PaletteSvg.White.as_color(),
                 sm_width=25.0, sm_height=25.0, **kwargs):
        super().__init__(position, color, sm_width=sm_width, sm_height=sm_height, **kwargs)
        self._position = position
        self._color = color
        self._is_dirty = True

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        self._position = Vec3(value)
        self._is_dirty = True

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value

    def _update(self):
        up = Vec3.unit_y()
        dir = (-self._position).normalised
        if dir == -Vec3.unit_y():
            up = Vec3.unit_x()
        elif dir == -Vec3.unit_y():
            up = Vec3.unit_z()
        self._view_mat = Mat4.look_at_gl(self._position, Vec3(0.0), up)
        self._proj_mat = Mat4.orthographic_gl(-self._sm_width / 2.0, self._sm_width / 2.0,
                                              -self._sm_height / 2.0, self._sm_height / 2.0,
                                              self._sm_near, self._sm_far)
        self._is_dirty = False

    @property
    def matrix(self):
        if self._is_dirty:
            self._update()
        return self._proj_mat * self._view_mat

    @property
    def view_matrix(self):
        if self._is_dirty:
            self._update()
        return self._view_mat

    @property
    def is_directional(self):
        return False

    @property
    def is_perspective(self):
        return False



import abc
import atexit
import enum
import re
import OpenGL.error

from .variable import glsl_types
from .. import gl
import uuid


class BindSemanticObject(metaclass=abc.ABCMeta):
    def __enter__(self):
        self._activate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._deactivate()

    @abc.abstractmethod
    def _activate(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _deactivate(self):
        raise NotImplementedError


class GpuObject:
    def __init__(self, gl_type, gl_id):
        self._id = gl_id
        self._uuid = uuid.uuid1()
        self._type = gl_type
        atexit.register(self.__delete)

    def __delete(self):
        try:
            # if the context is alive, delete the resource from GPU
            self._delete()
        except OpenGL.error.NullFunctionError as error:
            # do nothing; context is not existing anymore
            pass
        except gl.GLError as error:
            # do nothing, context doesn't exists anymore
            if error.err != 1282:
                raise error

    @abc.abstractmethod
    def _delete(self):
        raise NotImplementedError

    @property
    def handle(self):
        return self._id

    @property
    def uuid(self):
        return self._uuid

    @property
    def gl_type(self):
        return self._type

    def is_valid(self):
        return self._id > 0

    def delete(self):
        self._delete()


@enum.unique
class DrawingMode(enum.Enum):
    Points = gl.GL_POINTS
    Lines = gl.GL_LINES
    LineStrip = gl.GL_LINE_STRIP
    LineLoop = gl.GL_LINE_LOOP
    Triangles = gl.GL_TRIANGLES
    TrianglesAdjacency = gl.GL_TRIANGLES_ADJACENCY
    TriangleStrip = gl.GL_TRIANGLE_STRIP
    TriangleStripAdjacency = gl.GL_TRIANGLE_STRIP_ADJACENCY
    TriangleFan = gl.GL_TRIANGLE_FAN


# Valid OpenGL buffer binding targets.
@enum.unique
class BufferBindingTarget(enum.Enum):
    ArrayBuffer = gl.GL_ARRAY_BUFFER  # Vertex attributes
    AtomicCounterBuffer = gl.GL_ATOMIC_COUNTER_BUFFER  # Atomic counter storage, requires GL version >= 4.2
    CopyReadBuffer = gl.GL_COPY_READ_BUFFER  # Buffer copy source, requires GL version >= 3.1
    CopyWriteBuffer = gl.GL_COPY_WRITE_BUFFER  # Buffer copy destination
    DispatchIndirectBuffer = gl.GL_DISPATCH_INDIRECT_BUFFER  # Indirect compute dispatch commands, requires GL version >= 4.3
    DrawIndirectBuffer = gl.GL_DRAW_INDIRECT_BUFFER  # Indirect command arguments
    ElementArrayBuffer = gl.GL_ELEMENT_ARRAY_BUFFER  # Vertex array indices
    PixelPackBuffer = gl.GL_PIXEL_PACK_BUFFER  # Pixel read target
    PixelUnpackBuffer = gl.GL_PIXEL_UNPACK_BUFFER  # Texture data source
    QueryBuffer = gl.GL_QUERY_BUFFER  # Query result buffer, requires GL version >= 4.4
    ShaderStorageBuffer = gl.GL_SHADER_STORAGE_BUFFER  # Read-write storage for shaders, requires GL version >= 4.3
    TextureBuffer = gl.GL_TEXTURE_BUFFER  # Texture data buffer, requires GL version >= 3.1
    TransformFeedbackBuffer = gl.GL_TRANSFORM_FEEDBACK_BUFFER  # Transform feedback buffer
    UniformBuffer = gl.GL_UNIFORM_BUFFER  # Uniform block storage, requires GL version >= 3.1


# Valid OpenGL buffer data storage.
# https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glBufferStorage.xhtml
@enum.unique
class BufferStorageFlag(enum.Flag):
    Dynamic = gl.GL_DYNAMIC_STORAGE_BIT
    MapRead = gl.GL_MAP_READ_BIT
    MapWrite = gl.GL_MAP_WRITE_BIT
    MapPersistent = gl.GL_MAP_PERSISTENT_BIT
    MapCoherent = gl.GL_MAP_COHERENT_BIT
    ClientStorage = gl.GL_CLIENT_STORAGE_BIT


@enum.unique
class AccessPolicy(enum.Enum):
    """
    Access policy for OpenGL buffer object.
    """
    ReadOnly = gl.GL_READ_ONLY
    WriteOnly = gl.GL_WRITE_ONLY
    ReadWrite = gl.GL_READ_WRITE


@enum.unique
class DataUsage(enum.Enum):
    """
    Usage pattern of stored data.
    https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glBufferData.xhtml

    Frequency of access:
        - Stream: the data store contents will be modified once and used at most a few times.
        - Static: the data store contents will be modified once and used many times.
        - Dynamic: the data store contents will be modified repeatedly and used many times.

    Nature of access:
        - Draw:
            the data store contents are modified by the application, and used as the source for GL drawing
            and image specification commands.

        - Read:
            the data store contents are modified by reading data from the GL, and used to return that data
            when queried by the application.

        - Copy:
            the data store contents are modified by reading data from the GL, and used as the source for GL
             drawing and image specification commands.
    """
    StreamDraw = gl.GL_STREAM_DRAW
    StreamRead = gl.GL_STREAM_READ
    StreamCopy = gl.GL_STREAM_COPY
    StaticDraw = gl.GL_STATIC_DRAW
    StaticRead = gl.GL_STATIC_READ
    StaticCopy = gl.GL_STATIC_COPY
    DynamicDraw = gl.GL_DYNAMIC_DRAW
    DynamicRead = gl.GL_DYNAMIC_READ
    DynamicCopy = gl.GL_DYNAMIC_COPY


class ShaderCodeParser:
    @classmethod
    def parse(self, code):
        pass

    @classmethod
    def preprocess(cls, code: str):
        uniforms = []
        attribs = []
        if code:
            cls.remove_comments(code)
            structs = cls.parse_defined_struct(code)
            uniforms = cls.parse_declarations(code, 'uniform', user_defined_types=structs)
            attribs = cls.parse_declarations(code, 'in')
        return code, uniforms, attribs

    @classmethod
    def remove_comments(cls, code: str) -> str:
        """
        Replace C-style comment from GLSL code string.

        Args:
            code (str): GLSL code string.

        Returns:
            GLSL code string without comments.
        """
        regex = re.compile(r"(\".*?\"|\'.*?\')|(/\*.*?\*/|//[^\r\n]*\n)", re.MULTILINE | re.DOTALL)

        return regex.sub(
            lambda matched: "" if matched.group(2) is not None else matched.group(1),
            code
        )

    @classmethod
    def parse_defined_struct(cls, code: str):
        struct_regex = re.compile(r"(?s)struct\s*(?P<sname>\w+)\s*\{(?P<sdef>.*?)\};", re.MULTILINE)
        struct_field_regex = re.compile(r"\s+(?P<type>\w+)\s+(?P<name>[\w,\[\]\n = \.$]+);")
        structs = {}
        for matched in re.finditer(struct_regex, code):
            sname = matched.group('sname')
            structs[sname] = []
            for field in re.finditer(struct_field_regex, matched.group('sdef')):
                var_type = field.group('type')
                var_name = field.group('name')
                structs[sname].append((var_type, var_name))

        return structs

    @classmethod
    def remove_version(cls, code: str) -> str:
        """
        Remove OpenGL version directive.

        Args:
            code (str): GLSL code string.

        Returns:
            GLSL code string with version directive removed.
        """
        regex = re.compile('\#\s*version[^\r\n]*\n', re.MULTILINE | re.DOTALL)
        return regex.sub('\n', code)

    @classmethod
    def parse_declarations(cls, code, qualifiers='', user_defined_types={}):
        """Extract declaraions of different types (type qualifiers) inside of shader.

        Note:
              Do NOT pass multiple qualifiers (not working for the moment).

        Args:
            code (str):
                Shader source code string.

            qualifier (str):
                GLSL type qualifiers. See https://www.khronos.org/opengl/wiki/Type_Qualifier_(GLSL).
                A string of qualifier(s) separated by comma.

        Returns:
            list of parsed declarations.
        """
        # todo: deal with multiple qualifiers.
        if qualifiers != '':
            variables = []
            # qualifiers = f"({'|'.join(list(map(str.strip, qualifiers.split(','))))})"
            regex = re.compile(f'{qualifiers}\s+(?P<type>\w+)\s+(?P<names>[\w,\[\]\n = \.$]+);')
            for matched in re.finditer(regex, code):
                var_type = matched.group('type')
                var_names = list(map(str.strip, matched.group('names').split(',')))

                if var_type not in glsl_types and var_type in user_defined_types:
                    user_defined = user_defined_types[var_type]
                    for var_name in var_names:
                        for field_type, field_name in user_defined:
                            variables.append((f'{var_name}.{field_name}', field_type))
                else:
                    for var_name in var_names:
                        variables.append((var_name, var_type))
            return variables
        else:
            return ''



from __future__ import annotations

import math
from dataclasses import dataclass

from ..math import Vec3


@dataclass
class BoundingBox:
    lower: Vec3 = Vec3(math.inf)
    upper: Vec3 = Vec3(-math.inf)

    @classmethod
    def empty(cls) -> BoundingBox:
        return cls()

    @property
    def is_valid(self) -> bool:
        return self.lower == Vec3(math.inf) and self.upper == Vec3(-math.inf)



import enum
import hashlib
import logging
from collections import namedtuple

from .resolver import default_resolver
from .image import Image
from .. import gl
from ..graphics.program import ShaderProgram
from ..graphics.shader import Shader, ShaderType
from ..graphics.material import MaterialData
from ..graphics.texture import TextureWrapMode, FilterMode, TextureKind, Texture
from ..scene import mesh


class AssetKind(enum.Enum):
    Image = 0
    Texture = 1
    Material = 2


PipelineRecord = namedtuple('PipelineRecord', ['pipeline', 'vs', 'ps'])


class NameManager:
    def __init__(self):
        self._name_map = {}

    def get_name(self, name):
        if name not in self._name_map:
            self._name_map[name] = 0
        else:
            self._name_map[name] += 1

        return f'{name}_{self._name_map[name]:03}'


class AssetManager:
    def __init__(self, resolver=default_resolver):
        self._resolver = resolver
        self._materials = {}
        self._textures = {}
        self._images = {}
        self._shaders = {}
        self._pipelines = {}
        self._models = {}

    def get_or_create_image(self, image_path):
        """
        Returns the requested image. Load the image if it doesn't exist.

        Args:
            image_path (str):

        Returns:
            Pillow.Image
        """
        path = self._resolver.resolve(image_path)
        if path not in self._images:
            logging.info(f'-- Create image {path}')
            image = Image.open(path)
            self._images[path] = image

        logging.info(f'Load image <{path}>')
        return self._images[path]

    def get_or_create_texture(self, image_path, kind=TextureKind.DiffuseMap, target=gl.GL_TEXTURE_2D,
                              wrap_mode=TextureWrapMode.Repeat, filter_mode=FilterMode.Linear):
        path = self._resolver.resolve(image_path)
        image = self.get_or_create_image(path)
        texture_name = f'{path}_{kind.name}'
        if texture_name not in self._textures:
            logging.info(f'-- Create texture with image <{path}>')
            texture = Texture(path, image, kind, target, wrap_mode, filter_mode)
            self._textures[texture_name] = texture

        logging.info(f'Load texture <{texture_name}>')
        return self._textures[texture_name]

    def get_or_create_material(self, name, ambient=(0.8, 0.8, 0.8), diffuse=(0.8, 0.8, 0.8), specular=(1.0, 1.0, 1.0),
                               shininess=1.0, ior=1.0, dissolve=1.0, illum=2, **kwargs) -> MaterialData:
        """
        Args:
            name (str): name of the material
            ambient (array of 3 elements): ambient color
            diffuse (array of 3 elements): diffuse color
            specular (array of 3 elements): specular color
            illum (int): illumination model
            shininess (float): shininess of specular component
            dissolve (float): opacity
            ior (float): refractive index
            **kwargs:
                diffuse_map_path (str):
                    Specifies where to find the texture used as diffuse map.
                bump_map_path (str):
                    Specifies where to find the texture used as bump map.
                normal_map_path (str):
                    Specifies where to find the texture used as normal map.
        Returns:

        """
        if name not in self._materials:
            logging.info(f'-- Create material <{name}>')
            diffuse_map_path = kwargs.get('diffuse_map_path', None)
            if diffuse_map_path is None:
                diffuse_map_path = self._resolver.resolve('textures/checker.png')
            else:
                diffuse_map_path = self._resolver.resolve(diffuse_map_path)

            bump_map_path = kwargs.get('bump_map_path')
            if bump_map_path is None:
                bump_map_path = self._resolver.resolve('textures/checker_bump.png')
            else:
                bump_map_path = self._resolver.resolve(bump_map_path)

            normal_map_path = kwargs.get('normal_map_path')
            if normal_map_path is None:
                normal_map_path = self._resolver.resolve('textures/checker_normal.png')
            else:
                normal_map_path = self._resolver.resolve(normal_map_path)

            diffuse_map = self.get_or_create_texture(diffuse_map_path)
            bump_map = self.get_or_create_texture(bump_map_path, TextureKind.BumpMap)
            normal_map = self.get_or_create_texture(normal_map_path, TextureKind.NormalMap)
            material = MaterialData(name, diffuse_map, bump_map, normal_map, ambient, diffuse, specular, shininess, ior,
                                    dissolve, illum)
            self._materials[name] = material

        logging.info(f'Load material <{name}>')
        return self._materials[name]

    def get_or_create_shader(self, shader: str, kind: ShaderType, is_file=False):
        _, shader = self._get_or_create_shader(shader, kind, is_file)
        return shader

    def _get_or_create_shader(self, shader: str, kind: ShaderType, is_file=False):
        key = ''
        if is_file:
            key = self._resolver.resolve(shader)
            if key not in self._shaders:
                logging.info(f'-- Create shader from file <{key}>')
                with open(key, 'rt') as file:
                    self._shaders[key] = Shader(kind, code=file.read(), origin=key, is_file=False)
        else:
            key = hashlib.sha256(shader.encode('utf-8')).hexdigest()
            if key not in self._shaders:
                logging.info(f'-- Create shader from src <{key}>')
                self._shaders[key] = Shader(kind, code=shader, origin=f'<string#{key}>', is_file=False)

        return key, self._shaders[key]

    def get_or_create_pipeline(self, name: str, vertex_shader: str = None, pixel_shader: str = None) -> PipelineRecord:
        """

        Args:
            name (str): Specifies the name of the pipline (shader program).
            vertex_shader (str): Path string or code string of a vertex shader.
            pixel_shader (str): Path string or code string of a vertex shader.

        Returns:
            ShaderProgram
        """
        if name not in self._pipelines:
            logging.info(f'-- Create pipeline <{name}>')

            if vertex_shader is not None and pixel_shader is not None:
                is_file_vs = ';' not in vertex_shader
                is_file_ps = ';' not in pixel_shader

                vs_name = hashlib.sha256(
                    vertex_shader.encode('utf-8')).hexdigest() if not is_file_vs else self._resolver.resolve(
                    vertex_shader)
                ps_name = hashlib.sha256(
                    pixel_shader.encode('utf-8')).hexdigest() if not is_file_ps else self._resolver.resolve(
                    pixel_shader)

                for n, r in self._pipelines.items():
                    if r.vs == vs_name and r.ps == ps_name:
                        logging.info(f'Shaders exist in pipeline <{n}>')
                        return self._pipelines[n].pipeline

            if vertex_shader is None:
                vs_name, vs = self._get_or_create_shader('shaders/default.vert', ShaderType.Vertex, True)
            else:
                is_file_vs = ';' not in vertex_shader
                vs_name, vs = self._get_or_create_shader(vertex_shader, ShaderType.Vertex, is_file_vs)

            if pixel_shader is None:
                ps_name, ps = self._get_or_create_shader('shaders/default.frag', ShaderType.Pixel, True)
            else:
                is_file_ps = ';' not in pixel_shader
                ps_name, ps = self._get_or_create_shader(pixel_shader, ShaderType.Fragment, is_file_ps)

            self._pipelines[name] = PipelineRecord(ShaderProgram(vs, ps), vs_name, ps_name)

        logging.info(f'Load pipeline <{name}>')
        return self._pipelines[name].pipeline

    def get_pipeline(self, uuid):
        if len(self._pipelines) > 0:
            for record in self._pipelines.values():
                if record.pipeline.uuid == uuid:
                    return record.pipeline
        return None

    def get_or_load_wavefront_obj(self, filepath: str, resolver=None) -> dict:
        """
        Load a model from a file path if it is not loaded yet.

        Args:
            resolver: Path resolver.
            filepath (str): Path string of a model file.

        Returns:
            Model
        """
        resolver = resolver if resolver is not None else self._resolver
        path = resolver.resolve(filepath)
        logging.info(f'-- Getting model <{path}>')
        if path not in self._models:
            logging.info(f'  -- Loading... <{path}>')
            from ..scene.loader.obj import WavefrontReader
            reader = WavefrontReader(path)
            self._models[path] = reader.read()
        return self._models[path]


default_asset_mgr: AssetManager = AssetManager()
default_name_mgr: NameManager = NameManager()



import ctypes
import enum
import sys

import numpy as np

from .util import DataUsage, BufferBindingTarget, GpuObject, BindSemanticObject
from .vertex_layout import VertexLayout
from .. import gl


class QueryTarget(enum.Enum):
    SamplesPassed = gl.GL_SAMPLES_PASSED
    AnySamplesPassed = gl.GL_ANY_SAMPLES_PASSED
    AnySamplesPassedConservative = gl.GL_ANY_SAMPLES_PASSED_CONSERVATIVE
    PrimitivesGenerated = gl.GL_PRIMITIVES_GENERATED
    TransformFeedbackPrimitivesWritten = gl.GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN
    TimeElapsed = gl.GL_TIME_ELAPSED


class Query(GpuObject):
    def __init__(self, target: QueryTarget):
        super().__init__(target.value, -1)

        if target not in QueryTarget:
            raise ValueError(f'{target} is not a valid OpenGL query target.')

        self._target = target.value
        self._id = gl.glGenQueries(1)[0]
        self._is_in_use = False

    def _delete(self):
        if self.is_valid():
            gl.glDeleteQueries(1, [self._id])

    @property
    def target(self) -> gl.Constant:
        return self._target

    def begin(self):
        gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE)
        gl.glDepthMask(gl.GL_FALSE)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glBeginQuery(self._target, self._id)
        self._is_in_use = True

    def end(self):
        gl.glEndQuery(self._target)
        gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)
        gl.glDepthMask(gl.GL_TRUE)
        self._is_in_use = False

    def result(self):
        # check availability of the query result
        if gl.glGetQueryObjectiv(self._id, gl.GL_QUERY_RESULT_AVAILABLE):
            return gl.glGetQueryObjectiv(self._id, gl.GL_QUERY_RESULT)

    def is_in_use(self):
        return self._is_in_use
